# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional

from ansys.mapdl.core.mapdl_types import MapdlInt


class ElementTable:
    def detab(
        self,
        elem="",
        lab="",
        v1="",
        v2="",
        v3="",
        v4="",
        v5="",
        v6="",
        **kwargs,
    ):
        """Modifies element table results in the database.

        APDL Command: DETAB

        Parameters
        ----------
        elem
            Element for which results are to be modified.  If ALL, modify all
            selected elements [ESEL] results.  If ELEM = P, graphical picking
            is enabled and all remaining command fields are ignored (valid only
            in the GUI).  A component name may also be substituted for ELEM.

        lab
            Label identifying results.  Valid labels are as defined with the
            ETABLE command.  Issue ETABLE,STAT to display labels and values.

        v1
            Value assigned to this element table result in the database.  If
            zero, a zero value will be assigned.  If blank, value remains
            unchanged.

        v2, v3, v4, . . . , v6
            Additional values (if any) assigned to consecutive element table
            columns.

        Notes
        -----
        Modifies element table [ETABLE] results in the database.  For example,
        DETAB,35,ABC,1000,2000,1000 assigns 1000, 2000, and 1000 to the first
        three table columns starting with label ABC for element 35.  Use the
        PRETAB command to list the current results.  After deleting a column of
        data using ETABLE,Lab,ERASE, the remaining columns of data are not
        shifted to compress the empty slot.  Therefore, the user must allocate
        null (blank) values for V1, V2...V6 for any ETABLE entries which have
        been deleted by issuing ETABLE,Lab,ERASE.  All data are stored in the
        solution coordinate system but will be displayed in the results
        coordinate system [RSYS].
        """
        command = f"DETAB,{elem},{lab},{v1},{v2},{v3},{v4},{v5},{v6}"
        return self.run(command, **kwargs)

    def esort(
        self,
        item: str = "",
        lab: str = "",
        order: MapdlInt = "",
        kabs: MapdlInt = "",
        numb: MapdlInt = "",
        **kwargs,
    ) -> Optional[str]:
        """Sorts the element table.

        APDL Command: ESORT

        Parameters
        ----------
        item
            Label identifying the item:
            ETAB - (currently the only Item available)

        lab
            element table label: Lab - Any user-defined label from
            the ETABLE command (input in the Lab field of the ETABLE
            command).

        order
            Order of sort operation:

            0 - Sort into descending order.

            1 - Sort into ascending order.

        kabs
            Absolute value key:

            0 - Sort according to real value.

            1 - Sort according to absolute value.

        numb
            Number of elements (element table rows) to be sorted in
            ascending or descending order (ORDER) before sort is
            stopped (remainder will be in unsorted sequence)
            (defaults to all elements).

        Notes
        -----
        The element table rows are sorted based on the column
        containing the Lab values. Use EUSORT to restore the original
        order. If ESORT is specified with PowerGraphics on
        [/GRAPHICS,POWER], then the nodal solution results listing
        [PRNSOL] will be the same as with the full graphics mode
        [/GRAPHICS,FULL].
        """
        command = f"ESORT,{item},{lab},{order},{kabs},{numb}"
        return self.run(command, **kwargs)

    def etable(
        self,
        lab: str = "",
        item: str = "",
        comp: str = "",
        option: str = "",
        **kwargs,
    ) -> Optional[str]:
        """Fills a table of element values for further processing.

        APDL Command: ETABLE

        The ETABLE command defines a table of values per element (the
        element table) for use in further processing. The element
        table is organized similar to spreadsheet, with rows
        representing all selected elements and columns consisting of
        result items which have been moved into the table (Item,Comp)
        via ETABLE. Each column of data is identified by a
        user-defined label (Lab) for listings and displays.

        After entering the data into the element table, you are not
        limited to merely listing or displaying your data (PLESOL,
        PRESOL, etc.). You may also perform many types of operations
        on your data, such as adding or multiplying columns (SADD,
        SMULT), defining allowable stresses for safety calculations
        (SALLOW), or multiplying one column by another (SMULT).  See
        Getting Started in the Basic Analysis Guide for more
        information.

        Various results data can be stored in the element table. For
        example, many items for an element are inherently
        single-valued (one value per element). The single-valued items
        include: SERR, SDSG, TERR, TDSG, SENE, SEDN, TENE, KENE, AENE,
        JHEAT, JS, VOLU, and CENT. All other items are multivalued
        (varying over the element, such that there is a different
        value at each node). Because only one value is stored in the
        element table per element, an average value (based on the
        number of contributing nodes) is calculated for multivalued
        items. Exceptions to this averaging procedure are FMAG and all
        element force items, which represent the sum only of the
        contributing nodal values.

        Two methods of data access can be used with the ETABLE
        command. The method you select depends upon the type of data
        that you want to store.  Some results can be accessed via a
        generic label (Component Name method), while others require a
        label and number (Sequence Number method).

        The Component Name method is used to access the General
        element data (that is, element data which is generally
        available to most element types or groups of element
        types). All of the single-valued items and some of the more
        general multivalued items are accessible with the Component
        Name method.  Various element results depend on the
        calculation method and the selected results location (AVPRIN,
        RSYS, LAYER, SHELL, and ESEL).

        Although nodal data is readily available for listings and
        displays (PRNSOL, PLNSOL) without using the element table, you
        may also use the Component Name method to enter these results
        into the element table for further "worksheet"
        manipulation. (See Getting Started in theBasic Analysis Guide
        for more information.) A listing of the General Item and Comp
        labels for the Component Name method is shown below.

        The Sequence Number method allows you to view results for data
        that is not averaged (such as pressures at nodes, temperatures
        at integration points, etc.), or data that is not easily
        described in a generic fashion (such as all derived data for
        structural line elements and contact elements, all derived
        data for thermal line elements, layer data for layered
        elements, etc.). A table illustrating the Items (such as LS,
        LEPEL, LEPTH, SMISC, NMISC, SURF, etc.) and corresponding
        sequence numbers for each element is shown in the Output Data
        section of each element description found in the Element
        Reference.

        Some element table data are reported in the results coordinate
        system.  These include all component results (for example, UX,
        UY, etc.; SX, SY, etc.). The solution writes component results
        in the database and on the results file in the solution
        coordinate system. When you issue the ETABLE command, these
        results are then transformed into the results coordinate
        system (RSYS) before being stored in the element table. The
        default results coordinate system is global Cartesian
        (RSYS,0).  All other data are retrieved from the database and
        stored in the element table with no coordinate transformation.

        Use the PRETAB, PLETAB, or ETABLE,STAT commands to display the
        stored table values. Issue ETABLE,ERAS to erase the entire
        table. Issue ETABLE,Lab,ERAS to erase a Lab column.

        The element table data option (Option) is not available for
        all output items.

        Parameters
        ----------
        lab
            Any unique user defined label for use in subsequent
            commands and output headings (maximum of eight characters
            and not a General predefined Item label). Defaults to an
            eight character label formed by concatenating the first
            four characters of the Item and Comp labels. If the same
            as a previous user label, this result item will be
            included under the same label. Up to 200 different labels
            may be defined. The following labels are predefined and
            are not available for user-defined labels: ``'REFL''`,
            ``'STAT'``, and ``'ERAS'``.  ``lab='REFL'`` refills all
            tables previously defined with the :meth:`etable` commands
            (not the CALC module commands) according to the latest
            ETABLE specifications and is convenient for refilling
            tables after the load step (SET) has been
            changed. Remaining fields will be ignored if
            ``Lab='REFL'``.  ``lab='STAT'`` displays stored table
            values.  ``lab='ERAS'`` erases the entire table.

        item
            Label identifying the item. General item labels are shown
            in the table below. Some items also require a component
            label. Character parameters may be used. ``item='eras'``
            erases a Lab column.

        comp
            Component of the item (if required). General component
            labels are shown in the table below. Character parameters
            may be used.

        option
            Option for storing element table data:

            * ``'MIN'`` - Store minimum element nodal value of the specified
              item component.
            * ``'MAX'`` - Store maximum element nodal value of the specified
              item component.
            * ``'AVG'`` - Store averaged element centroid value of the
              specified item component (default).

        Examples
        --------
        Print the volume of individual elements.

        >>> mapdl.clear()
        >>> output = mapdl.input(examples.vmfiles['vm6'])
        >>> mapdl.post1()
        >>> label = 'MYVOLU'
        >>> mapdl.etable(label, 'VOLU')
        >>> print(mapdl.pretab(label))
        PRINT ELEMENT TABLE ITEMS PER ELEMENT
           *****ANSYS VERIFICATION RUN ONLY*****
             DO NOT USE RESULTS FOR PRODUCTION
          ***** POST1 ELEMENT TABLE LISTING *****
            STAT     CURRENT
            ELEM     XDISP
               1  0.59135E-001
               2  0.59135E-001
               3  0.59135E-001
        ...

        """
        command = f"ETABLE,{lab},{item},{comp},{option}"
        return self.run(command, **kwargs)

    def eusort(self, **kwargs) -> Optional[str]:
        """Restore original order of the element table.

        APDL Command: EUSORT

        Notes
        -----
        Changing the selected element set [ESEL] also restores the original
        element order.

        Examples
        --------
        >>> mapdl.post1()
        >>> mapdl.eusort()
        'ELEMENT SORT REMOVED'

        """
        return self.run("EUSORT", **kwargs)

    def pletab(self, itlab="", avglab="", **kwargs):
        """Displays element table items.

        APDL Command: PLETAB

        Parameters
        ----------
        itlab
            User-defined label, as specified with the ETABLE command, of item
            to be displayed.

        avglab
            Averaging operation:

            NOAV - Do not average element items at common nodes (default).

            AVG - Average the element items at common nodes.

        Notes
        -----
        Displays items stored in the table defined with the ETABLE command for
        the selected elements.  For display purposes, items are assumed to be
        constant over the element and assigned to each of its nodes.  Contour
        display lines (lines of constant value) are determined by linear
        interpolation within each element from the nodal values.  These nodal
        values have the option of being averaged (values are averaged at a node
        whenever two or more elements connect to the same node) or not averaged
        (discontinuous).  The discontinuity between contour lines of adjacent
        elements is an indication of the gradient across elements.

        Portions of this command are not supported by PowerGraphics
        [/GRAPHICS,POWER].
        """
        command = f"PLETAB,{itlab},{avglab}"
        return self.run(command, **kwargs)

    def plls(self, labi="", labj="", fact="", kund="", viewup="", **kwargs):
        """Displays element table items as contoured areas along elements.

        APDL Command: PLLS

        Parameters
        ----------
        labi
            Label of element table item (ETABLE) for node I magnitude.

        labj
            Label of element table item for node J magnitude.

        fact
            Scale factor for display (defaults to 1).  A negative scaling
            factor may be used to invert the display.

        kund
            Undisplaced shape key:

            0 - Display selected items on undeformed shape.

            1 - Display selected items on deformed shape.

        viewup
            View Up key:

            0 - Ignore the view-up (/VUP) vector when calculating trapezoid orientation
                (default).

            1 - Use the view-up (/VUP) vector to calculate trapezoid orientation.

        Notes
        -----
        Displays selected items (such as shears and moments) as a contoured
        area (trapezoid) display along line elements and 2-D axisymmetric shell
        elements (such as shear and moment diagrams).  Three sides of the
        trapezoid are formed by the element (one side) and lines at nodes I and
        J of length proportional to the item magnitude and displayed normal to
        the element and the viewing direction (the two parallel sides).

        When ViewUP = 1, the trapezoid is oriented within the plane created by
        the element and the global Cartesian coordinate system reference
        orientation (/VUP or view up) vector. In this case, the program does
        not perform the calculation involving the element and view direction.

        Portions of this command are not supported by PowerGraphics
        [/GRAPHICS,POWER].
        """
        command = f"PLLS,{labi},{labj},{fact},{kund},{viewup}"
        return self.run(command, **kwargs)

    def pretab(
        self,
        lab1="",
        lab2="",
        lab3="",
        lab4="",
        lab5="",
        lab6="",
        lab7="",
        lab8="",
        lab9="",
        **kwargs,
    ):
        """Print the element table items.

        APDL Command: PRETAB

        Parameters
        ----------
        lab1, lab2, lab3, ... , lab9
            Print selected items.  Valid labels are ``""`` or any
            label as specified with the ETABLE command.  Convenience
            labels may be used for Lab1 to select groups of labels (10
            labels maximum): GRP1 for first 10 stored items; GRP2 for
            items 11 to 20; GRP3 for items 21 to 30; GRP4 for items 31
            to 40; GRP5 for items 41 to 50.  Run ``etable("stat")``
            command to list stored item order.  If all labels are
            blank, print first 10 stored items (GRP1).

        Notes
        -----
        Prints the items stored in the table defined with the
        :func:`etable() <ansys.mapdl.core.Mapdl.etable>` command.
        Item values will be listed for the selected elements in the
        sorted sequence [ESORT].  :func:`force()
        <ansys.mapdl.core.Mapdl.force>` can be used to define which
        component of the nodal load is to be used (static, damping,
        inertia, or total).

        Examples
        --------
        Print out element displacement results.

        >>> mapdl.etable("values", "U", "X")
        STORE VALUES   FROM ITEM=U    COMP=X     FOR ALL SELECTED ELEMENTS

        Now print out the table.

        >>> mapdl.pretab().splitlines()[:4]
        ['PRINT ELEMENT TABLE ITEMS PER ELEMENT',
         '       1   0.1073961540E-005  0.1073961540E-005',
         '       2   0.3156317304E-005  0.3156317304E-005',
         '       3   0.5125435148E-005  0.5125435148E-005']

        Note that can access the array directly from APDL:

        >>> mapdl.get_array("elem", 1, "ETAB", "values")
        array([1.07396154e-06, 3.15631730e-06, 5.12543515e-06, ...,
               5.41204700e-06, 3.33649806e-06, 1.13836132e-06])

        """
        return self.run(
            f"PRETAB,{lab1},{lab2},{lab3},{lab4},{lab5},{lab6},{lab7},{lab8},{lab9}",
            **kwargs,
        )

    def sabs(self, key="", **kwargs):
        """Specifies absolute values for element table operations.

        APDL Command: SABS

        Parameters
        ----------
        key
            Absolute value key:

            0 - Use algebraic values in operations.

            1 - Use absolute values in operations.

        Notes
        -----
        Causes absolute values to be used in the SADD, SMULT, SMAX, SMIN, and
        SSUM operations.
        """
        command = f"SABS,{key}"
        return self.run(command, **kwargs)

    def sadd(self, labr="", lab1="", lab2="", fact1="", fact2="", const="", **kwargs):
        """Forms an element table item by adding two existing items.

        APDL Command: SADD

        Parameters
        ----------
        labr
            Label assigned to results.  If same as existing label, the existing
            values will be overwritten by these results.

        lab1
            First labeled result item in operation.

        lab2
            Second labeled result item in operation (may be blank).

        fact1
            Scale factor applied to Lab1. A (blank) or '0' entry defaults to
            1.0.

        fact2
            Scale factor applied to Lab2. A (blank) or '0' entry defaults to
            1.0.

        const
            Constant value.

        Notes
        -----
        Forms a labeled result (see ETABLE command) for the selected elements
        by adding two existing labeled result items according to the operation:

        LabR = (FACT1 x Lab1) + (FACT2 x Lab2) + CONST

        May also be used to scale results for a single labeled result item.  If
        absolute values are requested [SABS,1], absolute values of Lab1 and
        Lab2 are used.
        """
        command = f"SADD,{labr},{lab1},{lab2},{fact1},{fact2},{const}"
        return self.run(command, **kwargs)

    def sallow(
        self,
        strs1="",
        strs2="",
        strs3="",
        strs4="",
        strs5="",
        strs6="",
        **kwargs,
    ):
        """Defines the allowable stress table for safety factor calculations.

        APDL Command: SALLOW

        Parameters
        ----------
        strs1, strs2, strs3, . . . , strs6
            Input up to six allowable stresses corresponding to the temperature
            points [TALLOW].

        Notes
        -----
        Defines the allowable stress table for safety factor calculations
        [SFACT,SFCALC].  Use the STAT command to list current allowable stress
        table.  Repeat SALLOW to zero table and redefine points (6 maximum).

        Safety factor calculations are not supported by PowerGraphics. Both the
        SALLOW and TALLOW commands must be used with the Full Model Graphics
        display method active.
        """
        command = f"SALLOW,{strs1},{strs2},{strs3},{strs4},{strs5},{strs6}"
        return self.run(command, **kwargs)

    def sexp(self, labr="", lab1="", lab2="", exp1="", exp2="", **kwargs):
        """Forms an element table item by exponentiating and multiplying.

        APDL Command: SEXP

        Parameters
        ----------
        labr
            Label assigned to results.  If same as existing label, the existing
            values will be overwritten by these results.

        lab1
            First labeled result item in operation.

        lab2
            Second labeled result item in operation (may be blank).

        exp1
            Exponent applied to Lab1.

        exp2
            Exponent applied to Lab2.

        Notes
        -----
        Forms a labeled result item (see ETABLE command) for the selected
        elements by exponentiating and multiplying two existing labeled result
        items according to the operation:

        ``LabR = (|Lab1|EXP1) x (|Lab2|EXP2)``

        Roots, reciprocals, and divides may also be done with this command.
        """
        command = f"SEXP,{labr},{lab1},{lab2},{exp1},{exp2}"
        return self.run(command, **kwargs)

    def sfact(self, type_="", **kwargs):
        """Allows safety factor or margin of safety calculations to be made.

        APDL Command: SFACT

        Parameters
        ----------
        type\\_
            Type of calculation:

            0 - No nodal safety factor or margin of safety calculations.

            1 - Calculate and store safety factors in place of nodal stresses.

            2 - Calculate and store margins of safety in place of nodal stresses.

        Notes
        -----
        Allows safety factor (SF) or margin of safety (MS) calculations to be
        made for the average nodal stresses according to:

        ``SF = SALLOW/|Stress|``

        ``MS = (SALLOW/|Stress|) -- 1.0``

        Calculations are done during the display, select, or sort operation (in
        the active coordinate system [RSYS]) with results stored in place of
        the nodal stresses.  Use the PRNSOL or PLNSOL command to display the
        results.

        The results are meaningful only for the stress (SIG1, SIGE,
        etc.) upon which SALLOW is based.  Nodal temperatures used are those
        automatically stored for the node.  Related commands are SFCALC,
        SALLOW, TALLOW.
        """
        command = f"SFACT,{type_}"
        return self.run(command, **kwargs)

    def sfcalc(self, labr="", labs="", labt="", type_="", **kwargs):
        """Calculates the safety factor or margin of safety.

        APDL Command: SFCALC

        Parameters
        ----------
        labr
            Label assigned to results.  If same as existing label, the existing
            values will be overwritten by these results.

        labs
            Labeled result item corresponding to the element stress.

        labt
            Labeled result item corresponding to the element temperature.

        type\\_
            Type of calculation:

            0 or 1 - Use safety factor (SF) calculation.

            2 - Use margin of safety (MS) calculation.

            3 - Use 1/SF calculation.

        Notes
        -----
        Calculates safety factor (SF) or margin of safety (MS) as described for
        the SFACT command for any labeled result item (see ETABLE command) for
        the selected elements.  Use the PRETAB or PLETAB command to display
        results.  Allowable element stress is determined from the SALLOW-TALLOW
        table [SALLOW, TALLOW].
        """
        command = f"SFCALC,{labr},{labs},{labt},{type_}"
        return self.run(command, **kwargs)

    def smax(self, labr="", lab1="", lab2="", fact1="", fact2="", **kwargs):
        """Forms an element table item from the maximum of two other items.

        APDL Command: SMAX

        Parameters
        ----------
        labr
            Label assigned to results.  If same as existing label, the existing
            values will be overwritten by these results.

        lab1
            First labeled result item in operation.

        lab2
            Second labeled result item in operation (may be blank).

        fact1
            Scale factor applied to Lab1. A (blank) or '0' entry defaults to
            1.0.

        fact2
            Scale factor applied to Lab2. A (blank) or '0' entry defaults to
            1.0.

        Notes
        -----
        Forms a labeled result item (see ETABLE command) for the selected
        elements by comparing two existing labeled result items according to
        the operation:

        LabR = (FACT1 x Lab1) cmx (FACT2 x Lab2)

        where "cmx" means "compare and save maximum."  If absolute values are
        requested [SABS,1], the absolute values of Lab1 and Lab2 are used.
        """
        command = f"SMAX,{labr},{lab1},{lab2},{fact1},{fact2}"
        return self.run(command, **kwargs)

    def smin(self, labr="", lab1="", lab2="", fact1="", fact2="", **kwargs):
        """Forms an element table item from the minimum of two other items.

        APDL Command: SMIN

        Parameters
        ----------
        labr
            Label assigned to results.  If same as existing label, the existing
            values will be overwritten by these results.

        lab1
            First labeled result item in operation.

        lab2
            Second labeled result item in operation (may be blank).

        fact1
            Scale factor applied to Lab1. A (blank) or '0' entry defaults to
            1.0.

        fact2
            Scale factor applied to Lab2. A (blank) or '0' entry defaults to
            1.0.

        Notes
        -----
        Forms a labeled result item (see ETABLE command) for the selected
        elements by comparing two existing labeled result items according to
        the operation:

        LabR = (FACT1 x Lab1) cmn (FACT2 x Lab2)

        where "cmn" means "compare and save minimum."  If absolute values are
        requested [SABS,1], the absolute values of Lab1 and Lab2 are used.
        """
        command = f"SMIN,{labr},{lab1},{lab2},{fact1},{fact2}"
        return self.run(command, **kwargs)

    def smult(self, labr="", lab1="", lab2="", fact1="", fact2="", **kwargs):
        """Forms an element table item by multiplying two other items.

        APDL Command: SMULT

        Parameters
        ----------
        labr
            Label assigned to results.  If same as existing label, the existing
            values will be overwritten by these results.

        lab1
            First labeled result item in operation.

        lab2
            Second labeled result item in operation (may be blank).

        fact1
            Scale factor applied to Lab1. A (blank) or '0' entry defaults to
            1.0.

        fact2
            Scale factor applied to Lab2. A (blank) or '0' entry defaults to
            1.0.

        Notes
        -----
        Forms a labeled result item (see ETABLE command) for the selected
        elements by multiplying two existing labeled result items according to
        the operation:

        LabR = (FACT1 x Lab1) x (FACT2 x Lab2)

        May also be used to scale results for a single labeled result item.  If
        absolute values are requested [SABS,1], the absolute values of Lab1 and
        Lab2 are used.
        """
        command = f"SMULT,{labr},{lab1},{lab2},{fact1},{fact2}"
        return self.run(command, **kwargs)

    def ssum(self, **kwargs):
        """Calculates and prints the sum of element table items.

        APDL Command: SSUM

        Notes
        -----
        Calculates and prints the tabular sum of each existing labeled result
        item [ETABLE] for the selected elements.  If absolute values are
        requested [SABS,1], absolute values are used.
        """
        command = f"SSUM,"
        return self.run(command, **kwargs)

    def tallow(
        self,
        temp1="",
        temp2="",
        temp3="",
        temp4="",
        temp5="",
        temp6="",
        **kwargs,
    ):
        """Defines the temperature table for safety factor calculations.

        APDL Command: TALLOW

        Parameters
        ----------
        temp1, temp2, temp3, . . . , temp6
            Input up to six temperatures covering the range of nodal
            temperatures.  Temperatures must be input in ascending order.

        Notes
        -----
        Defines the temperature table for safety factor calculations [SFACT,
        SALLOW].  Use STAT command to list current temperature table.  Repeat
        TALLOW command to zero table and redefine points (6 maximum).

        Safety factor calculations are not supported by PowerGraphics. Both the
        SALLOW and TALLOW commands must be used with the Full Model Graphics
        display method active.
        """
        command = f"TALLOW,{temp1},{temp2},{temp3},{temp4},{temp5},{temp6}"
        return self.run(command, **kwargs)

    def vcross(
        self,
        labxr="",
        labyr="",
        labzr="",
        labx1="",
        laby1="",
        labz1="",
        labx2="",
        laby2="",
        labz2="",
        **kwargs,
    ):
        """Forms element table items from the cross product of two vectors.

        APDL Command: VCROSS

        Parameters
        ----------
        labxr, labyr, labzr
            Label assigned to X, Y, and Z-component of resultant vector.

        labx1, laby1, labz1
            X, Y, and Z-component of first vector label.

        labx2, laby2, labz2
            X, Y, and Z-component of second vector label.

        Notes
        -----
        Forms labeled result items for the selected element from the cross
        product of two vectors:

        {LabXR, LabYR, LabZR} = {LabX1, LabY1, LabZ1} X {LabX2, LabY2, LabZ2}

        Data must be in a consistent coordinate system.  Labels are those
        associated with the ETABLE command.
        """
        command = f"VCROSS,{labxr},{labyr},{labzr},{labx1},{laby1},{labz1},{labx2},{laby2},{labz2}"
        return self.run(command, **kwargs)

    def vdot(
        self,
        labr="",
        labx1="",
        laby1="",
        labz1="",
        labx2="",
        laby2="",
        labz2="",
        **kwargs,
    ):
        """Forms an element table item from the dot product of two vectors.

        APDL Command: VDOT

        Parameters
        ----------
        labr
            Label assigned to dot product result.

        labx1, laby1, labz1
            X, Y, and Z-component of first vector label.

        labx2, laby2, labz2
            X, Y, and Z-component of second vector label.

        Notes
        -----
        Forms labeled result items for the selected element from the dot
        product of two vectors:

        LabR = {LabX1, LabY1, LabZ1} : :  {LabX2, LabY2, LabZ2}

        Data must be in a consistent coordinate system.  Labels are those
        associated with the ETABLE command.
        """
        command = f"VDOT,{labr},{labx1},{laby1},{labz1},{labx2},{laby2},{labz2}"
        return self.run(command, **kwargs)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class ConstraintEquations:
    def ce(
        self,
        neqn="",
        const="",
        node1="",
        lab1="",
        c1="",
        node2="",
        lab2="",
        c2="",
        node3="",
        lab3="",
        c3="",
        **kwargs,
    ):
        """Defines a constraint equation relating degrees of freedom.

        APDL Command: CE

        Parameters
        ----------
        neqn
            Set equation reference number:

            n - Arbitrary set number.

            HIGH - The highest defined constraint equation number. This option is especially
                   useful when adding nodes to an existing set.

            NEXT - The highest defined constraint equation number plus one. This option
                   automatically numbers coupled sets so that existing sets are
                   not modified.

        const
            Constant term of equation.

        node1
            Node for first term of equation.  If -NODE1, this term is deleted
            from the equation.

        lab1
            Degree of freedom label for first term of equation.  Structural
            labels:  UX, UY, or UZ (displacements); ROTX, ROTY, or ROTZ
            (rotations, in radians).  Thermal labels: TEMP, TBOT, TE2, TE3, . .
            ., TTOP (temperature).  Electric labels:  VOLT (voltage).  Magnetic
            labels:  MAG (scalar magnetic potential); AX, AY, or AZ (vector
            magnetic potentials). Diffusion label: CONC (concentration).

        c1
            Coefficient for first node term of equation.  If zero, this term is
            ignored.

        node2, lab2, c2
            Node, label, and coefficient for second term.

        node3, lab3, c3
            Node, label, and coefficient for third term.

        Notes
        -----
        Repeat the CE command to add additional terms to the same equation.  To
        change only the constant term, repeat the command with no node terms
        specified.  Only the constant term can be changed during solution, and
        only with the CECMOD command.

        Linear constraint equations may be used to relate the degrees of
        freedom of selected nodes in a more general manner than described for
        nodal coupling [CP].  The constraint equation is of the form:

        where U(I) is the degree of freedom (displacement, temperature, etc.)
        of term (I).  The following example is a set of two constraint
        equations, each containing three terms:

        0.0 = 3.0* (1 UX) + 3.0* (4 UX) + (-2.0)* (4 ROTY)

        2.0 = 6.0* (2 UX) + 10.0* (4 UY) + 1.0* (3 UZ)

        The first unique degree of freedom in the equation is eliminated in
        terms of all other degrees of freedom in the equation.  A unique degree
        of freedom is one which is not specified in any other constraint
        equation, coupled node set, specified displacement set, or master
        degree of freedom set.  It is recommended that the first term of the
        equation be the degree of freedom to be eliminated.  The first term of
        the equation cannot contain a master degree of freedom, and no term can
        contain coupled degrees of freedom.  The same degree of freedom may be
        specified in more than one equation but care must be taken to avoid
        over-specification (over-constraint).

        The degrees of freedom specified in the equation (i.e., UX, UY, ROTZ,
        etc.) must also be included in the model (as determined from the
        element types [ET]).  Also, each node in the equation must be defined
        on an element (any element type containing that degree of freedom will
        do).

        For buckling and modal analyses, the constant term of the equation will
        not be taken into account (that is, CONST is always zero).

        Note that under certain circumstances a constraint equation generated
        by CE may be modified during the solution. See Program Modification of
        Constraint Equations for more information.
        """
        command = f"CE,{neqn},{const},{node1},{lab1},{c1},{node2},{lab2},{c2},{node3},{lab3},{c3}"
        return self.run(command, **kwargs)

    def cecyc(
        self,
        lowname="",
        highname="",
        nsector="",
        hindex="",
        tolerance="",
        kmove="",
        kpairs="",
        **kwargs,
    ):
        """Generates the constraint equations for a cyclic symmetry analysis

        APDL Command: CECYC

        Parameters
        ----------
        lowname
            Name of a component for the nodes on the low angle edge of the
            sector.  Enclosed in single quotes.

        highname
            Name of a component for the nodes on the high angle edge of the
            sector.  Enclosed in single quotes.

        nsector
            Number of sectors in the complete 360 degrees.

        hindex
            Harmonic index to be represented by this set of constraint
            equations.  If Hindex is -1, generate constraint equations for
            static cyclic symmetry.  If HIndex is -2, generate constraint
            equations for static cyclic asymmetry.

        tolerance
            A positive tolerance is an absolute tolerance (length units), and a
            negative tolerance is a tolerance relative to the local element
            size.

        kmove
            0

            0 - Nodes are not moved.

            1 - HIGHNAME component nodes are moved to match LOWNAME component nodes exactly.

        kpairs
            0

            0 - Do not print paired nodes

            1 - Print table of paired nodes

        Notes
        -----
        The analysis can be either modal cyclic symmetry or static cyclic
        symmetry.

        The pair of nodes for which constraint equations are written are
        rotated into CSYS,1.
        """
        command = f"CECYC,{lowname},{highname},{nsector},{hindex},{tolerance},{kmove},{kpairs}"
        return self.run(command, **kwargs)

    def cedele(self, neqn1="", neqn2="", ninc="", nsel="", **kwargs):
        """Deletes constraint equations.

        APDL Command: CEDELE

        Parameters
        ----------
        neqn1, neqn2, ninc
            Delete constraint equations from NEQN1 to NEQN2 (defaults to NEQN1)
            in steps of NINC (defaults to 1).  If NEQN1 = ALL, NEQN2  and NINC
            will be ignored all constraint equations will be deleted.

        nsel
            Additional node selection control:

            ANY  - Delete equation set if any of the selected nodes are in the set (default).

            ALL  - Delete equation set only if all of the selected nodes are in the set.
        """
        command = f"CEDELE,{neqn1},{neqn2},{ninc},{nsel}"
        return self.run(command, **kwargs)

    def ceintf(
        self,
        toler="",
        dof1="",
        dof2="",
        dof3="",
        dof4="",
        dof5="",
        dof6="",
        movetol="",
        **kwargs,
    ):
        """Generates constraint equations at an interface.

        APDL Command: CEINTF

        Parameters
        ----------
        toler
            Tolerance about selected elements, based on a fraction of the
            element dimension (defaults to 0.25 (25%)).  Nodes outside the
            element by more than the tolerance are not accepted as being on the
            interface.

        dof1, dof2, dof3, . . . , dof6
            Degrees of freedom for which constraint equations are written.
            Defaults to all applicable DOFs.  DOF1 accepts ALL as a valid
            label, in which case the rest are ignored (all DOFs are applied).

        movetol
            The allowed "motion" of a node (see Note below).  This distance is
            in terms of the element coordinates (-1.0 to 1.0).  A typical value
            is 0.05.  Defaults to 0 (do not move).  MoveTol must be less than
            or equal to TOLER.

        Notes
        -----
        This command can be used to "tie" together two regions with dissimilar
        mesh patterns by generating constraint equations that connect the
        selected nodes of one region to the selected elements of the other
        region.  At the interface between regions, nodes should be selected
        from the more dense mesh region, A, and the elements selected from the
        less dense mesh region, B.  The degrees of freedom of region A nodes
        are interpolated with the corresponding degrees of freedom of the nodes
        on the region B elements, using the shape functions of the region B
        elements.  Constraint equations are then written that relate region A
        and B nodes at the interface.

        The MoveTol field lets the nodes in the previously mentioned region A
        change coordinates when slightly inside or outside the elements of
        region B.  The change in coordinates causes the nodes of region A to
        assume the same surface as the nodes associated with the elements of
        region B.   The constraint equations that relate the nodes at both
        regions of the interface are then written.

        Solid elements with six degrees of freedom should only be interfaced
        with other six degree-of-freedom elements.  The region A nodes should
        be near the region B elements.   A location tolerance based on the
        smallest region B element length may be input.  Stresses across the
        interface are not necessarily continuous.  Nodes in the interface
        region should not have specified constraints.

        Use the CPINTF command to connect nodes by coupling instead of
        constraint equations.  Use the EINTF command to connect nodes by line
        elements.  See also the NSEL and ESEL commands for selecting nodes and
        elements.  See the Mechanical APDL Theory Reference for a description
        of 3-D space used to determine if a node will be considered by this
        command.

        As an alternative to the CEINTF command, you can use contact elements
        and the internal multipoint constraint (MPC) algorithm to tie together
        two regions having dissimilar meshes. See Solid-Solid and Shell-Shell
        Assemblies for more information.
        """
        command = f"CEINTF,{toler},{dof1},{dof2},{dof3},{dof4},{dof5},{dof6},{movetol}"
        return self.run(command, **kwargs)

    def celist(self, neqn1="", neqn2="", ninc="", option="", **kwargs):
        """Lists the constraint equations.

        APDL Command: CELIST

        Parameters
        ----------
        neqn1, neqn2, ninc
            List constraint equations from NEQN1 to NEQN2 (defaults to NEQN1)
            in steps of NINC (defaults to 1).  If NEQN1 = ALL (default), NEQN2
            and NINC are ignored and all constraint equations are listed.

        option
            Options for listing constraint equations:

            ANY - List equation set if any of the selected nodes are in the set (default). Only
                  externally-generated constraint equations are listed.

            ALL - List equation set only if all of the selected nodes are in the set. Only
                  externally-generated constraint equations are listed.

            INTE - List internally-generated constraint equations that are associated with MPC-
                   based contact. Constraint equations are listed only if all
                   the nodes in the set are selected.

            CONV - Convert internal constraint equations to external constraint equations.
                   Internal constraint equations are converted only if all of
                   the nodes in the set are selected.

        Notes
        -----
        This command is valid in any processor. However, the INTE and CONV
        options are only valid in the Solution processor after a SOLVE command
        has been issued.
        """
        command = f"CELIST,{neqn1},{neqn2},{ninc},{option}"
        return self.run(command, **kwargs)

    def cerig(
        self,
        maste="",
        slave="",
        ldof="",
        ldof2="",
        ldof3="",
        ldof4="",
        ldof5="",
        **kwargs,
    ):
        """Defines a rigid region.

        APDL Command: CERIG

        Parameters
        ----------
        maste
            Retained (or master) node for this rigid region.  If MASTE = P,
            then graphical picking of the master and slave nodes is enabled
            (first node picked will be the master node, and subsequent nodes
            picked will be slave  nodes), and subsequent fields are ignored
            (valid only in GUI).

        slave
            Removed (or slave) node for this rigid region.  If ALL, slave nodes
            are all selected nodes.

        ldof
            Degrees of freedom associated with equations:

            ALL - All applicable degrees of freedom (default).  If 3-D, generate 6 equations
                  based on UX, UY, UZ, ROTX, ROTY, ROTZ;  if 2-D, generate 3
                  equations based on UX, UY, ROTZ.

            UXYZ - Translational degrees of freedom.  If 3-D, generate 3 equations based on the
                   slave nodes' UX, UY, and UZ DOFs and the master node's UX,
                   UY, UZ, ROTX, ROTY, and ROTZ DOFs;  if 2-D, generate 2
                   equations based on the slave nodes UX and UY DOFs and the
                   master nodes UX, UY, and ROTZ DOFs.  No equations are
                   generated for the rotational coupling.

            RXYZ - Rotational degrees of freedom.  If 3-D, generate 3 equations based on ROTX,
                   ROTY, ROTZ;  if 2-D, generate 1 equation based on ROTZ.  No
                   equations are generated for the translational coupling.

            UX - Slave translational UX degree of freedom only.

            UY - Slave translational UY degree of freedom only.

            UZ - Slave translational UZ degree of freedom only.

            ROTX - Slave rotational ROTX degree of freedom only.

            ROTY - Slave rotational ROTY degree of freedom only.

            ROTZ - Slave rotational ROTZ degree of freedom only.

        ldof2, ldof3, ldof4, ldof5
            Additional degrees of freedom.  Used only if more than one degree
            of freedom required and Ldof is not ALL, UXYZ, or RXYZ.

        Notes
        -----
        Defines a rigid region (link, area or volume) by automatically
        generating constraint equations to relate nodes in the region.  Nodes
        in the rigid region must be assigned a geometric location before this
        command is used.  Also, nodes must be connected to elements having the
        required degree of freedom set (see Ldof above).  Generated constraint
        equations are based on small deflection theory.  Generated constraint
        equations are numbered beginning from the highest previously defined
        equation number (NEQN) plus 1.  Equations, once generated, may be
        listed [CELIST] or modified [CE] as desired.  Repeat CERIG command for
        additional rigid region equations.

        This command will generate the constraint equations needed for defining
        rigid lines in 2-D or 3-D space.  Multiple rigid lines relative to a
        common point are used to define a rigid area or a rigid volume.  In 2-D
        space, with Ldof = ALL, three equations are generated for each pair of
        constrained nodes.  These equations define the three rigid body motions
        in global Cartesian space, i.e., two in-plane translations and one in-
        plane rotation.  These equations assume the X-Y plane to be the active
        plane with UX, UY, and ROTZ degrees of freedom available at each node.
        Other types of equations can be generated with the appropriate Ldof
        labels.

        Six equations are generated for each pair of constrained nodes in 3-D
        space (with Ldof = ALL).  These equations define the six rigid body
        motions in global Cartesian space.  These equations assume that UX, UY,
        UZ, ROTX, ROTY, and ROTZ degrees of freedom are available at each node.

        The UXYZ label allows generating a partial set of rigid region
        equations.  This option is useful for transmitting the bending moment
        between elements having different degrees of freedom at a node.  With
        this option only two of the three equations are generated for each pair
        of constrained nodes in 2-D space.  In 3-D space, only three of the six
        equations are generated.  In each case the rotational coupling
        equations are not generated.  Similarly, the RXYZ label allows
        generating a partial set of equations with the translational coupling
        equations omitted.

        Applying this command to a large number of slave nodes may result in
        constraint equations with a large number of coefficients. This may
        significantly increase the peak memory required during the process of
        element assembly. If real memory or virtual memory is not available,
        consider reducing the number of slave nodes.

        Note that under certain circumstances the constraint equations
        generated by CERIG may be modified during the solution. See Program
        Modification of Constraint Equations for more information.

        As an alternative to the CERIG command, you can define a similar type
        of rigid region using contact elements and the internal multipoint
        constraint (MPC) algorithm. See Surface-Based Constraints for more
        information.

        CERIG cannot be deleted using CEDELE,ALL and then regenerated in the
        second or higher load steps if the LSWRITE  and LSSOLVE procedure is
        used. CERIG writes constraint equations directly into load step files.
        Deleting constraint equations (CEDELE,ALL) cannot always maintain the
        consistency among load steps.
        """
        command = f"CERIG,{maste},{slave},{ldof},{ldof2},{ldof3},{ldof4},{ldof5}"
        return self.run(command, **kwargs)

    def cesgen(self, itime="", inc="", nset1="", nset2="", ninc="", **kwargs):
        """Generates a set of constraint equations from existing sets.

        APDL Command: CESGEN

        Parameters
        ----------
        itime, inc
            Do this generation operation a total of ITIMEs, incrementing all
            nodes in the existing sets by INC each time after the first.  ITIME
            must be >1 for generation to occur.

        nset1, nset2, ninc
            Generate sets from sets beginning with NSET1 to NSET2 (defaults to
            NSET1) in steps of NINC (defaults to 1).  If NSET1 is negative,
            NSET2 and NINC are ignored and the last ``|NSET1|`` sets (in sequence
            from maximum set number) are used as the sets to be repeated.

        Notes
        -----
        Generates additional sets of constraint equations (with same labels)
        from existing sets.  Node numbers between sets may be uniformly
        incremented.
        """
        command = f"CESGEN,{itime},{inc},{nset1},{nset2},{ninc}"
        return self.run(command, **kwargs)

    def rbe3(self, master="", dof="", slaves="", wtfact="", **kwargs):
        """Distributes the force/moment applied at the master node to a set  of

        APDL Command: RBE3
        slave nodes, taking into account the geometry of the slave nodes as
        well as weighting factors.

        Parameters
        ----------
        master
            Node at which the force/moment to be distributed will be applied.
            This node must be associated with an element for the master node to
            be included in the DOF solution.

        dof
            Refers to the master node degrees of freedom to be used in
            constraint equations. Valid labels are: UX, UY, UZ, ROTX, ROTY,
            ROTZ, UXYZ, RXYZ, ALL

        slaves
            The name of an array parameter that contains a list of slave nodes.
            Must specify the starting index number. ALL can be used for
            currently selected set of nodes. The slave nodes may not be
            colinear, that is, not be all located on the same straight line
            (see Notes below).

        wtfact
            The name of an array parameter that contains a list of weighting
            factors corresponding to each slave node above. Must have the
            starting index number. If not specified, the weighting factor for
            each slave node defaults to 1.

        Notes
        -----
        The force is distributed to the slave nodes proportional to the
        weighting factors. The moment is distributed as forces to the slaves;
        these forces are proportional to the distance from the center of
        gravity of the slave nodes times the weighting factors.  Only the
        translational degrees of freedom of the slave nodes are used for
        constructing the constraint equations. Constraint equations are
        converted to distributed forces/moments on the slave nodes during
        solution.

        RBE3 creates constraint equations such that the motion of the master is
        the average of the slaves. For the rotations, a least-squares approach
        is used to define the "average rotation" at the master from the
        translations of the slaves. If the slave nodes are colinear, then one
        of the master rotations that is parallel to the colinear direction can
        not be determined in terms of the translations of the slave nodes.
        Therefore, the associated moment component on the master node in that
        direction can not be transmitted. When this case occurs, a warning
        message is issued and the constraint equations created by RBE3 are
        ignored.

        Applying this command to a large number of slave nodes may result in
        constraint equations with a large number of coefficients. This may
        significantly increase the peak memory required during the process of
        element assembly. If real memory or virtual memory is not available,
        consider reducing the number of slave nodes.

        As an alternative to the RBE3 command, you can apply a similar type of
        constraint using contact elements and the internal multipoint
        constraint (MPC) algorithm. See Surface-based Constraints for more
        information.

        This command is also valid in SOLUTION.
        """
        command = f"RBE3,{master},{dof},{slaves},{wtfact}"
        return self.run(command, **kwargs)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class DynamicOptions:
    def alphad(self, value="", **kwargs):
        """Defines the mass matrix multiplier for damping.

        APDL Command: ALPHAD

        Parameters
        ----------
        value
            Mass matrix multiplier for damping.

        Notes
        -----
        This command defines the mass matrix multiplier α used to form the
        viscous damping matrix [C] = α[M] where [M] is the mass matrix.

        Values of  α may also be input as a material property (use the ALPD
        label on the MP command). If ALPD is included, the ALPD value is added
        to the ALPHAD value as appropriate (see Damping Matrices in the
        Mechanical APDL Theory Reference). Damping is not used in the static
        (ANTYPE,STATIC) or buckling (ANTYPE,BUCKLE) analyses.

        This command is also valid in PREP7.
        """
        command = f"ALPHAD,{value}"
        return self.run(command, **kwargs)

    def betad(self, value="", **kwargs):
        """Defines the stiffness matrix multiplier for damping.

        APDL Command: BETAD

        Parameters
        ----------
        value
            Stiffness matrix multiplier for damping.

        Notes
        -----
        This command defines the stiffness matrix multiplier β used to form the
        viscous damping matrix [C] = β [K] where [K] is the stiffness matrix.

        Values of : β may also be input as a material property (use the BETD
        label on the MP command).  If BETD is included, the BETD value is added
        to the BETAD value as appropriate (see Damping Matrices in the
        Mechanical APDL Theory Reference).  Damping is not used in the static
        (ANTYPE,STATIC) or buckling (ANTYPE,BUCKLE) analyses.

        This command is also valid in PREP7.
        """
        command = f"BETAD,{value}"
        return self.run(command, **kwargs)

    def dmprat(self, ratio="", **kwargs):
        """Sets a constant modal damping ratio.

        APDL Command: DMPRAT

        Parameters
        ----------
        ratio
            Modal damping ratio (for example, 2% is input as 0.02).

        Notes
        -----
        Sets a constant damping ratio for use in the mode-superposition
        transient (ANTYPE,TRANS) or harmonic (ANTYPE,HARMIC) analysis and the
        spectrum (ANTYPE,SPECTR) analysis.

        This command is also valid in PREP7.
        """
        command = f"DMPRAT,{ratio}"
        return self.run(command, **kwargs)

    def dmpstr(self, coeff="", **kwargs):
        """Sets a constant structural damping coefficient.

        APDL Command: DMPSTR

        Parameters
        ----------
        coeff
            Structural damping coefficient.

        Notes
        -----
        Sets a constant structural (or hysteretic) damping coefficient for use
        in harmonic (ANTYPE,HARMIC) analyses (FULL, MSUP, and VT) and modal
        analyses (ANTYPE,MODAL with MODOPT,UNSYM, DAMP or QRDAMP).

        Note that for structures with multiple materials, MP,DMPR can also be
        used to specify constant structural material damping on a per material
        basis. Note that if both DMPSTR and MP,DMPR are specified, the damping
        effects are additive.

        Caution:: : DMPSTR adds the damping contribution as gK, whereas MP,DMPR
        adds the contribution on a per-material basis as 2gK. For more
        information, see Damping Matrices in the Mechanical APDL Theory
        Reference.

        This command is also valid in PREP7.
        """
        command = f"DMPSTR,{coeff}"
        return self.run(command, **kwargs)

    def frqscl(self, scaling="", **kwargs):
        """Turns on automatic scaling of the entire mass matrix and frequency

        APDL Command: FRQSCL
        range for modal analyses using the Block Lanczos, PCG Lanczos, or
        Supernode mode extraction method.

        Parameters
        ----------
        scaling

            Off
                Do not use automatic scaling of the mass matrix and frequency range.

            On
                Use automatic scaling of the mass matrix and frequency range.

        Notes
        -----
        Use this command to deactivate or force activation of automatic scaling
        of the entire mass matrix and frequency range for modal analyses where
        the entire mass matrix is significantly different (i.e., orders of
        magnitude difference) than the entire stiffness matrix (for example,
        due to the particular unit system being used).  Where the mass matrix
        is significantly smaller compared to the stiffness matrix, the
        eigenvalues will tend to approach very large numbers (>10e12), making
        the Block Lanczos, PCG Lanczos, or Supernode mode extraction method
        less efficient and more likely to miss modes.

        ANSYS uses scaling (if appropriate) by default. However, you can issue
        FRQSCL,ON to force the entire mass matrix and frequency range to be
        scaled to bring the stiffness and mass matrices closer together in
        terms of orders of magnitude, improving efficiency and reducing the
        likelihood of missed modes.  The resulting eigenvalues are then
        automatically scaled back to the original system. If you are using
        micro MKS units, where the density is typically very small compared to
        the stiffness, you may want to issue FRQSCL,ON to force scaling on.

        If the stiffness and mass are on the same scale, FRQSCL,ON has no
        effect.

        This command is available only for modal analyses using the Block
        Lanczos, PCG Lanczos, or Supernode mode extraction method (MODOPT,LANB,
        LANPCG, or SNODE).

        This command is not valid and has no effect when used in conjunction
        with the MSAVE,ON command in a modal analysis with the PCG Lanczos mode
        extraction method.
        """
        command = f"FRQSCL,{scaling}"
        return self.run(command, **kwargs)

    def harfrq(self, freqb="", freqe="", logopt="", freqarr="", toler="", **kwargs):
        """Defines the frequency range in a harmonic analysis.

        APDL Command: HARFRQ

        Parameters
        ----------
        freqb
            Frequency (Hz) at the beginning of the FREQB to FREQE range (if
            FREQE > FREQB).  If FREQE is blank, the solution is done only at
            frequency FREQB (the central frequency of octave bands, when LogOpt
            = OB1, OB2, OB3, OB6, OB12 or OB24).

        freqe
            Frequency at end of this range.  Solutions are done at an interval
            of (FREQE-FREQB) / NSBSTP, ending at FREQE.  No solution is done at
            the beginning of the frequency range.  NSBSTP is input via the
            NSUBST command.  See the EXPSOL command documentation for expansion
            pass solutions.

        --
            Reserved.

        logopt
            Logarithm frequency span. Solutions are done at an interval of
            (log(FREQE) - log(FREQB)) / (NSBSTP-1), (NSBSTP>1). The central
            frequency or beginning frequency is used for NSBSTP = 1. Valid
            values are:

            OB1
                Octave band.

            OB2
                1/2 octave band.

            OB3
                1/3 octave band.

            OB6
                1/6 octave band.

            OB12
                1/12 octave band.

            OB24
                1/24 octave band.

            LOG
                General logarithm frequency span.

        freqarr
            An array containing frequency values (Hz). Combined with the
            tolerance argument, Toler, these values are merged with values
            calculated based on the specifications from FREQB, FREQE, and
            LogOpt, as well NSBSTP on the NSUBST command and Clust on the HROUT
            command. Enclose the array name in percent (%) signs (for example,
            HARFRQ,,,,,%arrname%). Use ``*DIM`` to define the array.

        toler
            Tolerance to determine if a user input frequency value in FREQARR
            is a duplicate and can be ignored. Two frequency values are
            considered duplicates if their difference is less than the
            frequency range multiplied by the tolerance. The default value is 1
            x 10-5.

        Notes
        -----
        Defines the frequency range for loads in the harmonic analysis
        (ANTYPE,HARMIC).

        Do not use this command for a harmonic ocean wave analysis (HROCEAN).

        When frequencies are user-defined, the array FREQARR must be one-
        dimensional and contain positive values. User-defined frequency input
        is not supported in the following cases:

        in a cyclic symmetry harmonic analysis

        when the Variational Technology method is used (Method = VT on the
        HROPT command)

        This command is also valid in PREP7.
        """
        command = f"HARFRQ,{freqb},{freqe},{logopt},{freqarr},{toler}"
        return self.run(command, **kwargs)

    def hrexp(self, angle="", **kwargs):
        """Specifies the phase angle for the harmonic analysis expansion pass.

        APDL Command: HREXP

        Parameters
        ----------
        angle
            Phase angle (degrees) for expansion pass.  If ALL (default), use
            both 0.0° (real) and 90.0° (imaginary) phase angles.

        Notes
        -----
        Specifies the phase angle where the expansion pass will be done for a
        harmonic mode-superposition expansion pass.

        For a specific angle, the following real solution is stored in the
        results (``*.rst``) file:

        If ANGLE is ALL, both the real and imaginary parts of the solution are
        stored in the results file.

        For more details about the solution equations, see Harmonic Analyses in
        the Mechanical APDL Theory Reference.

        This command is ignored if the HROPT command has been issued with
        Method = VT or Method = VTRU.

        This command is also valid in PREP7.
        """
        command = f"HREXP,{angle}"
        return self.run(command, **kwargs)

    def hrocean(self, type_="", nphase="", **kwargs):
        """Perform the harmonic ocean wave procedure (HOWP).

        APDL Command: HROCEAN

        Parameters
        ----------
        type\\_
            Specifies how to include ocean wave information in a harmonic
            analysis:

            HARMONIC
                Performs a harmonic analysis using both real and imaginary load vectors
                calculated via the harmonic ocean wave procedure (HOWP).
                This behavior is the default. This option performs a
                harmonic analysis running at a frequency determined by
                the wave period (specified via OCTABLE command input).

            STATIC
                Performs a static analysis using both real and imaginary load vectors
                (calculated via HOWP). This option works by performing a
                harmonic analysis running at a frequency of 0.0.

            OFF
                Deactivates a previously activated HOWP and performs a standard harmonic
                analysis.

        nphase
            Positive number specifying the number of phases to calculate
            forces. This value must be at least 8. The default value is 20.

        Notes
        -----
        The HROCEAN command applies ocean wave information (obtained via the
        OCDATA and OCTABLE commands) in a harmonic analysis (ANTYPE,HARMIC) as
        real and imaginary forces.

        You can apply only one ocean load at a time.

        The applied frequency in the harmonic (Type = HARMONIC) analysis is
        based on the wave period input on the OCTABLE command (and not on
        HARFRQ command input, which cannot be used). Phase-shift input on the
        OCTABLE command is ignored.

        HOWP does not generate a damping matrix. If you require a damping
        matrix, you must add it separately.

        The command applies to regular wave types only (Airy with one wave
        component, Wheeler with one wave component, Stokes, and stream
        function). Irregular wave types are not supported. For information
        about wave types, see Hydrodynamic Loads in the Mechanical APDL Theory
        Reference.

        The program calculates the forces on each load component of each
        element at NPHASE solutions, spread evenly over one wave cycle. Then,
        the minimum and maximum, and the phase between them, are calculated.
        The command uses the resulting information to generate the real and
        imaginary loads.

        HOWP cannot be used with stress stiffening.

        HOWP works with the full harmonic analysis method (HROPT,FULL) only.

        For more information, see Harmonic Ocean Wave Procedure (HOWP) in the
        Mechanical APDL Theory Reference.

        This command is also valid in PREP7.
        """
        command = f"HROCEAN,{type_},{nphase}"
        return self.run(command, **kwargs)

    def hropt(self, method="", maxmode="", minmode="", mcout="", damp="", **kwargs):
        """Specifies harmonic analysis options.

        APDL Command: HROPT

        Parameters
        ----------
        method
            Solution method for the harmonic analysis:

            AUTO
                Automatically select the most efficient method. Either the FULL method or the
                Variational Technology method is selected depending on the
                model. (default method).

            FULL
                Full method.

            MSUP
                Mode-superposition method.

            VT
                Variational Technology method (based on FULL harmonic algorithm).

            VTPA
                Variational Technology perfect absorber method (based on FULL harmonic
                algorithm).

            VTRU
                Variational Technology reuse method (based on FULL harmonic algorithm).

        maxmode
            Largest mode number to be used to calculate the response (for
            Method = MSUP only).  Defaults to the highest mode calculated in
            the preceding modal analysis.

        minmode
            Smallest mode number to be used (for Method = MSUP only).  Defaults
            to 1.

        mcout
            Modal coordinates output key (valid only for the mode superposition
            method MSUP):

            NO
                No output of modal coordinates (default).

            YES
                Output modal coordinates to the text file jobname.MCF.

        damp
            Damping mode for frequency-dependent material properties (valid
            only for the Variational Technology Method VT).

            Hysteretic
                Not proportional to the frequency.

            Viscous
                Proportional to the frequency (default).

        Notes
        -----
        Specifies the method of solution for a harmonic analysis
        (ANTYPE,HARMIC).  If used in SOLUTION, this command is valid only
        within the first load step. See the product restrictions indicated
        below.

        For cyclic symmetry mode-superposition harmonic solutions, MAXMODE and
        MINMODE are ignored.

        To include residual vectors in your mode-superposition harmonic
        analysis, specify RESVEC,ON.

        This command is also valid in PREP7.

        Distributed ANSYS Restriction: The VTRU method is not supported.
        """
        command = f"HROPT,{method},{maxmode},{minmode},{mcout},{damp}"
        return self.run(command, **kwargs)

    def hrout(self, reimky="", clust="", mcont="", **kwargs):
        """Specifies the harmonic analysis output options.

        APDL Command: HROUT

        Parameters
        ----------
        reimky
            Real/Imaginary print key:

            ON
                Print complex displacements as real and imaginary components (default).

            OFF
                Print complex displacements as amplitude and phase angle (degrees).

        clust
            Cluster option (for HROPT,MSUP):

            OFF
                Uniform spacing of frequency solutions (default).

            ON
                Cluster frequency solutions about natural frequencies.

        mcont
            Mode contributions key (for HROPT,MSUP):

            OFF
                No print of mode contributions at each frequency (default).

            ON
                Print mode contributions at each frequency.

        Notes
        -----
        Specifies the harmonic analysis (ANTYPE,HARMIC) output options.  If
        used in SOLUTION, this command is valid only within the first load
        step. OUTPR,NSOL must be specified to print mode contributions at each
        frequency.

        This command is ignored if the HROPT command has been issued with
        Method = VT, VTPA, or VTRU. Displacements are not available at expanded
        frequencies with these solution methods.

        For cyclic symmetry mode-superposition harmonic solutions, the cluster
        option is not available.

        This command is also valid in PREP7.
        """
        command = f"HROUT,{reimky},{clust},{mcont}"
        return self.run(command, **kwargs)

    def lvscale(self, fact="", ldstep="", **kwargs):
        """Scales the load vector for mode-superposition analyses.

        APDL Command: LVSCALE

        Parameters
        ----------
        fact
            Scale factor applied to both the real and imaginary (if they exist)
            components of the load vector. Defaults to 0.0.

        ldstep
            Specifies the load step number from the modal analysis
            (MODCONT,ON). It corresponds to the load vector number. Defaults to
            1. The maximum value is 240.

        Notes
        -----
        Specifies the scale factor for the load vector that was created in a
        modal (ANTYPE,MODAL) analysis.  Applies only to the mode-superposition
        transient analysis (ANTYPE,TRANS), mode-superposition harmonic analysis
        (ANTYPE,HARMIC), random vibration analysis (ANTYPE,SPECTR with
        SPOPT,PSD), and multiple point response spectrum analysis
        (ANTYPE,SPECTR with SPOPT,MPRS).  For PSD and MPRS analyses, LVSCALE is
        only applicable for pressure loading.

        The LVSCALE command supports tabular boundary conditions (%TABNAME_X%)
        for FACT input values only as a function of time in the mode-
        superposition transient (ANTYPE,TRANS) or as a function of frequency in
        mode-superposition harmonic (ANTYPE,HARMIC).

        MPC contact generates constraint equations that can include constant
        terms (included on the right-hand side of the system equation). The
        LVSCALE command scales the constant terms.

        In mode-superposition transient and harmonic analyses, all of the load
        vectors need to be scaled in the first load step. Use a zero scale
        factor if they are not actually used in this first load step. :
        Similarly, in random vibration and multipoint response spectrum
        analyses, all of the load vectors need to be scaled in the first
        participation factor calculation (PFACT). : Use a zero scale factor if
        they are not actually used for the first input table.

        This command is also valid in PREP7.
        """
        command = f"LVSCALE,{fact},{ldstep}"
        return self.run(command, **kwargs)

    def mascale(self, massfact="", **kwargs):
        """Activates scaling of the entire system matrix.

        APDL Command: MASCALE

        Parameters
        ----------
        massfact
           Scaling factor (> 0) for the mass matrix. Default = 1.0.

        Notes
        -----
        This command is supported in the first load step of the analysis only.
        The following features are not affected by the scaling:

        * Ocean loading
        * Steady-state rolling SSTATE

        The mass-related information (mass, center of mass, and mass
        moments of inertia) printed in the mass summary is based on
        unscaled mass properties.
        """
        return self.run(f"MASCALE,{massfact}", **kwargs)

    def mdamp(self, stloc="", v1="", v2="", v3="", v4="", v5="", v6="", **kwargs):
        """Defines the damping ratios as a function of mode.

        APDL Command: MDAMP

        Parameters
        ----------
        stloc
            Starting location in table for entering data.  For example, if
            STLOC = 1, data input in the V1 field applies to the first constant
            in the table.  If STLOC = 7, data input in the V1 field applies to
            the seventh constant in the table, etc.  Defaults to the last
            location filled + 1.

        v1, v2, v3, . . . , v6
            Data assigned to six locations starting with STLOC.  If a value is
            already in this location, it will be redefined.  Blank values for
            V2 to V6 leave the corresponding previous value unchanged.

        Notes
        -----
        Defines the damping ratios as a function of mode.  Table position
        corresponds to mode number.  These ratios are added to the DMPRAT
        value, if defined.  Use STAT command to list current values.  Applies
        to the mode-superposition harmonic (ANTYPE,HARMIC), the mode-
        superposition linear transient dynamic (ANTYPE,TRANS), and the spectrum
        (ANTYPE,SPECTR) analyses.  Repeat MDAMP command for additional
        constants (10000 maximum).

        MDAMP can also be defined in a substructure analysis using component
        mode synthesis with fixed-interface method (ANTYPE,SUBSTR with
        CMSOPT,FIX and SEOPT,,,3). The damping ratios are added to the diagonal
        of the reduced damping matrix as explained in Component Mode Synthesis
        (CMS).

        This command is also valid in PREP7.
        """
        command = f"MDAMP,{stloc},{v1},{v2},{v3},{v4},{v5},{v6}"
        return self.run(command, **kwargs)

    def mdplot(self, function="", dmpname="", scale="", **kwargs):
        """Plots frequency-dependent modal damping coefficients calculated by

        APDL Command: MDPLOT
        DMPEXT.

        Parameters
        ----------
        function
            Function to display.

            d_coeff
                Damping coefficient

            s_coeff
                Squeeze coefficient

            d_ratio
                Damping ratio

            s_ratio
                Squeeze stiffness ratio

        dmpname
            Array parameter name where damping information is stored. Defaults
            to d_damp.

        scale
            Indicates whether to perform a linear or a double logarithmic plot.

            LIN
                Perform a linear plot. Default

            LOG
                Perform a double logarithmic plot.

        Notes
        -----
        See Thin Film Analysis for more information on thin film analyses.
        """
        command = f"MDPLOT,{function},{dmpname},{scale}"
        return self.run(command, **kwargs)

    def midtol(self, key="", tolerb="", resfq="", **kwargs):
        """Sets midstep residual criterion values for structural transient

        APDL Command: MIDTOL
        analyses.

        Parameters
        ----------
        key
            Midstep residual criterion activation key.

            ON or 1
                Activate midstep residual criterion in a structural transient analysis
                (default).

            OFF or 0
                Deactivate midstep residual criterion in a structural transient analysis.

            STAT
                List the current midstep residual criterion setting.

        tolerb
            Midstep residual tolerance or reference value for bisection.
            Defaults to 100 times the TOLER setting of the CNVTOL command.

        resfq
            Key to use response frequency computation along with midstep
            residual criterion for automatic time stepping (AUTOTS,ON).

            OFF or 0
                Do not calculate response frequency and do not consider it in the automatic
                time stepping (default).

            ON or 1
                Calculate response frequency and consider it in the automatic time stepping.

        Notes
        -----
        When TOLERB is input as a tolerance value (TOLERB > 0), the typical
        force and/or moment from the regular time step is used in the midstep
        residual force and/or moment comparison.

        In a structural transient analysis, the suggested tolerance range of
        TOLERB (TOLERB > 0) is as follows:

        If the structural transient analysis is elastic and linear, and the
        load is constant or changes slowly, use a smaller value of TOLERB to
        achieve an accurate solution. If the analysis involves large amounts of
        energy dissipation, such as elastic-plastic material, TOLERB can be
        larger. If the analysis includes contact or rapidly varying loads, a
        smaller value of TOLERB should be used if high frequency response is
        important; otherwise, a larger value of TOLERB may be used to enable
        faster convergence with larger time step sizes.

        For more information on how the midstep criterion is used by the
        program, see Midstep Residual for Structural Dynamic Analysis in the
        Mechanical APDL Theory Reference.

        This command is also valid in PREP7.
        """
        command = f"MIDTOL,{key},{tolerb},{resfq}"
        return self.run(command, **kwargs)

    def modcont(self, mlskey="", enforcedkey="", **kwargs):
        """Specify additional modal analysis options.

        APDL Command: MODCONT

        Parameters
        ----------
        mlskey
            Multiple load step key:

            OFF
                Perform the modal analysis (compute the eigenvalues and the load vector) for
                each load step. (default)

            ON
                Perform the modal analysis (compute the eigenvalues and the load vector) only
                for the first load step; form the load vector for each
                subsequent load step (without repeating the eigenvalue
                calculations) and write all load vectors to the Jobname.MODE
                file for downstream mode-superposition analyses.

        enforcedkey
            Enforced motion key:

            OFF
                Do not calculate enforced static modes. (default)

            ON
                Calculate enforced static modes and write them to the Jobname.MODE file.

        Notes
        -----
        Specifies additional modal analysis (ANTYPE,MODAL) options.

        Use the LVSCALE command to apply the desired load in a mode-
        superposition transient or harmonic analysis.

        The maximum number of load vectors that can be used in the downstream
        mode-superposition transient or harmonic analysis is: 240.

        Generation of multiple loads (MLSkey = ON) is supported by the Block
        Lanczos, PCG Lanczos, Supernode, Subspace, Unsymmetric, and QR damped
        modal methods.

        The enforced motion calculation (EnforcedKey = ON) is supported by the
        Block Lanczos and Supernode mode extraction methods.
        """
        command = f"MODCONT,{mlskey},{enforcedkey}"
        return self.run(command, **kwargs)

    def modseloption(
        self, dir1="", dir2="", dir3="", dir4="", dir5="", dir6="", **kwargs
    ):
        """APDL Command: MODSELOPTION

        Parameters
        ----------
        dir1, dir2, dir3, dir4, dir5, dir6
            Selection of the direction to be expanded.

            For ``modeselmethod=effm`` on the MXPAND command, the
            directions correspond to the global Cartesian directions,
            i.e. 1=X, 2=Y, 3=Z, 4=ROTX, 5=ROTY, and 6=ROTZ. If dir1 = YES,
            then any mode in this direction is expanded if its modal
            effective mass divided by the total mass (modal effective mass
            ratio) is greater than SIGNIF on the MXPAND command. If
            dir1=NO, then the specified direction is not considered as a
            criterion for expansion. If dir1 is given a numerical decimal
            value, modes in that direction are selected (starting from the
            ones with the largest modal effective mass ratios to the
            smallest) until the sum of their modal effective mass ratio
            equals this requested threshold.

            For ModeSelMethod = MODC on the MXPAND command, dir1
            corresponds to the first input spectrum, dir2 to the second,
            etc. (i.e. for multiple spectrum inputs; the actual directions
            correspond to their respective SED directions). If dir1=YES,
            then any mode in this spectrum is ex- panded if its mode
            coefficient divided by the largest mode coefficient is greater
            than SIGNIF on the MXPAND command. If dir1=NO, then the
            specified direction is not considered as a criterion for
            expansion.

        Notes
        -----
        This command is only applicable when a mode selection method is defined
        (ModeSelMethod on the MXPAND command). See Using Mode Selection in the
        Mechanical APDL Structural Analysis Guide for more details.

        If a numerical value is specified for a direction, the significance
        threshold (SIGNIF on the MXPAND command) is ignored for the selection
        of the modes in this direction.

        If a mode is determined to be expanded in any of the 6 directions, it
        will be expanded in the .MODE file. : Otherwise, the mode will not be
        expanded.

        The default behavior is to consider all directions for expansion.
        """
        command = f"MODSELOPTION,{dir1},{dir2},{dir3},{dir4},{dir5},{dir6}"
        return self.run(command, **kwargs)

    def modopt(
        self,
        method="",
        nmode="",
        freqb="",
        freqe="",
        cpxmod="",
        nrmkey="",
        modtype="",
        blocksize="",
        freqmod="",
        **kwargs,
    ):
        """Specifies modal analysis options.

        APDL Command: MODOPT

        Parameters
        ----------
        method
            Mode-extraction method to be used for the modal analysis.

            LANB
                Block Lanczos

            LANPCG
                PCG Lanczos

            SNODE
                Supernode modal solver

            SUBSP
                Subspace algorithm

            UNSYM
                Unsymmetric matrix

            DAMP
                Damped system

            QRDAMP
                Damped system using QR algorithm

            VT
                Variational Technology

        nmode
            The number of modes to extract. The value can depend on the value
            supplied for Method. NMODE has no default and must be specified. If
            Method = LANB, LANPCG, or SNODE, the number of modes that can be
            extracted can equal the DOFs in the model after the application of
            all boundary conditions.

        freqb
            The beginning, or lower end, of the frequency range of interest.

        freqe
            The ending, or upper end, of the frequency range of interest (in
            Hz). The default for Method = SNODE is described below. The default
            for all other methods is to calculate all modes, regardless of
            their maximum frequency.

        cpxmod
            Complex eigenmode key. (Valid only when Method = QRDAMP or Method =
            UNSYM).

            AUTO
                Determine automatically if the eigensolutions are real or complex and output
                them accordingly. This is the default for Method = UNSYM.
                Not supported for Method = QRDAMP.

            ON or CPLX
                Calculate and output complex eigenmode shapes.

            OFF or REAL
                Do not calculate complex eigenmode shapes. This is required if a mode-
                superposition analysis is intended after the modal
                analysis for Method = QRDAMP. This is the default for
                this method.

        nrmkey
            Mode shape normalization key:

            OFF
                Normalize the mode shapes to the mass matrix (default).

            ON
                Normalize the mode shapes to unity instead of to the mass matrix.  If a
                subsequent spectrum or mode-superposition analysis is planned,
                the mode shapes should be normalized to the mass matrix
                (Nrmkey = OFF).

        modtype
            Type of modes calculated by the eigensolver. Only applicable to the
            unsymmetric eigensolver.

            Blank
                Right eigenmodes. This value is the default.

            BOTH
                Right and left eigenmodes. The left eigenmodes are written to Jobname.LMODE.
                This option must be activated if a mode-superposition
                analysis is intended.

        blocksize
            The block vector size to be used with the Block Lanczos or
            Subspace eigensolver (used only when Method = LANB or
            SUBSP). BlockSize must be an integer value between 0 and
            16. When BlockSize = zero or blank, the code decides the
            block size internally (normally, a value of 8 is used for
            LANB and a value of 6 is used for SUBSP).  Typically,
            higher BlockSize values are more efficient under each of
            the following conditions:

            - When running in out-of-core mode and there is not enough
              physical memory to buffer all of the files written by
              the Block Lanczos or Subspace eigensolver (and thus, the
              time spent doing I/O is considerable).
            - Many modes are requested (>100).
            - Higher-order solid elements dominate the model.

            The memory usage only slightly increases as BlockSize is
            increased. It is recommended that you use a value
            divisible by 4 (4, 8, 12, or 16).

        freqmod
            The specified frequency when the solved eigenvalues are no
            longer frequencies (for example, the model has the Floquet
            periodic boundary condition). In a modal analysis, the
            Floquet periodic boundary condition (body load FPBC) is
            only valid for the acoustic elements FLUID30, FLUID220,
            and FLUID221.

        Notes
        -----
        Specifies modal analysis (ANTYPE,MODAL) options. Additional options
        used only for the Supernode (SNODE) eigensolver are specified by the
        SNOPTION command. Additional options used only for the Subspace (SUBSP)
        eigensolver are specified by the SUBOPT command. If Method = LANPCG,
        ANSYS automatically switches to the PCG solver internally for this
        modal analysis. You can further control the efficiency of the PCG
        solver with the PCGOPT and EQSLV commands.

        For models that involve a non-symmetric element stiffness matrix, as in
        the case of a contact element with frictional contact, the QRDAMP
        eigensolver (MODOPT, QRDAMP) extracts modes in the modal subspace
        formed by the eigenmodes from the symmetrized eigenproblem. The QRDAMP
        eigensolver symmetrizes the element stiffness matrix on the first pass
        of the eigensolution, and in the second pass, eigenmodes are extracted
        in the modal subspace of the first eigensolution pass. For such non-
        symmetric eigenproblems, you should verify the eigenvalue and eigenmode
        results using the non-symmetric matrix eigensolver  (MODOPT,UNSYM).

        The DAMP and QRDAMP options cannot be followed by a subsequent spectrum
        analysis. The UNSYM method supports spectrum analysis when
        eigensolutions are real.

        This command is also valid in PREP7.

        Distributed ANSYS Restriction: The VT extraction method is not
        supported in Distributed ANSYS. All other extraction methods are
        supported. However, PCG Lanczos, SUBSP, UNSYM, DAMP, and QRDAMP are the
        only distributed eigensolvers that will run a fully distributed
        solution. The Block Lanczos and Supernode eigensolvers are not
        distributed eigensolvers; therefore, you will not see the full
        performance improvements with these methods that you would with a fully
        distributed solution.
        """
        command = f"MODOPT,{method},{nmode},{freqb},{freqe},{cpxmod},{nrmkey},{modtype},{blocksize},,,,{freqmod}"
        return self.run(command, **kwargs)

    def mxpand(
        self,
        nmode="",
        freqb="",
        freqe="",
        elcalc="",
        signif="",
        msupkey="",
        modeselmethod="",
        **kwargs,
    ):
        """Specifies the number of modes to expand and write for a modal or

        APDL Command: MXPAND
        buckling analysis.

        Parameters
        ----------
        nmode
            Number of modes or array name (enclosed in percent signs) to expand
            and write. If blank or ALL, expand and write all modes within the
            frequency range specified. If -1, do not expand and do not write
            modes to the results file during the analysis. If an array name is
            input, the array must contain 1 for the expanded modes and zero
            otherwise, where the array index corresponds to the mode number. To
            specify an array containing the individual modes to expand, enclose
            the array name in percent (%) signs (for example,
            MXPAND,%arrname%). Use the ``*DIM`` command to define the array.

        freqb
            Beginning, or lower end, of frequency range of interest. If FREQB
            and FREQE are both blank, expand and write the number of modes
            specified without regard to the frequency range. Defaults to the
            entire range.

        freqe
            Ending, or upper end, of frequency range of interest.

        elcalc
            Element calculation key:

            NO
                Do not calculate element results, reaction forces, and energies (default).

            YES
                Calculate element results, reaction forces, energies, and the nodal degree of
                freedom solution.

        signif
            Expand only those modes whose significance level exceeds the SIGNIF
            threshold (only applicable when ModeSelMethod is defined).

        msupkey
            Element result superposition key:

            NO
                Do not write element results to the mode file Jobname.MODE.

            YES
                Write element result to the mode file for use in the expansion pass of a
                subsequent mode-superposition PSD, transient, or harmonic
                analysis (default if Elcalc = YES and the mode shapes are
                normalized to the mass matrix).

        modeselmethod
            Methods for mode selection (not supported for complex
            eigensolvers):

            blank
                No mode selection is performed (default).

            MODM
                The mode selection is based on the modal effective masses.

            MODC
                The mode selection is based on the mode coefficients.

            DDAM
                The mode selection is based on DDAM procedure (see Mode Selection Based on DDAM
                Procedure in the Mechanical APDL Structural Analysis Guide
                for more information). This option is applicable only to
                DDAM spectrum analysis.

        Notes
        -----
        Specifies the number of modes to expand and write over a frequency
        range for a modal (ANTYPE,MODAL) or buckling (ANTYPE,BUCKLE) analysis.
        If used in SOLUTION, this command is valid only within the first load
        step.

        There is no limit on the number of expanded modes (NMODE). However,
        there is a limit on the maximum number of modes used via the ``*GET,,MODE``
        command, mode combinations, and the MDAMP command.

        With MSUPkey = YES, the computed element results (Elcalc = YES) are
        written to Jobname.MODE for use in subsequent downstream mode-
        superposition analyses, including harmonic, transient, and PSD
        analyses. This significantly reduces computation time for the
        combination or expansion passes. For limitations, see Option: Number of
        Modes to Expand (MXPAND) in the Mechanical APDL Structural Analysis
        Guide.

        If a mode selection method (ModeSelMethod) is defined, only the
        selected modes will be expanded. See Using Mode Selection in the
        Mechanical APDL Structural Analysis Guide for more details about the
        procedure.

        For array input (NMODE), the array must be dimensioned to be the size
        of the number of modes extracted (NMODE on the MODOPT command). A value
        of 1 in the array indicates the mode is to be expanded, and a value of
        0 indicates not to expand the mode. For the DAMP modal solution, the
        modes are in pairs, so be sure to verify that both modes of a pair have
        the same value. (For example, if modes #3 and #4 are a pair, indices 3
        and 4 in the array should have the same value, 0 or 1.)

        For linear perturbation modal analyses, you must set both Elcalc and
        MSUPkey to YES so that the downstream stress expansion pass can produce
        a solution consistent with the linear or nonlinear base (static or full
        transient) analysis. The prestressed nonlinear element history (saved
        variables) is accessible only in the first and second phases of the
        linear perturbation. The downstream MSUP or PSD analysis can only reuse
        the nonlinear information contained in the Jobname.MODE file that is
        generated in the linear perturbation.

        In a Distributed ANSYS analysis, you must issue MXPAND to specify the
        number of modes to expand when computing the modes and mode shapes. In
        a Distributed ANSYS run, MXPAND cannot be issued in an expansion pass
        (EXPASS).

        This command is also valid in PREP7.
        """
        command = f"MXPAND,{nmode},{freqb},{freqe},{elcalc},{signif},{msupkey},{modeselmethod}"
        return self.run(command, **kwargs)

    def qrdopt(self, reusekey="", symmeth="", cmccoutkey="", **kwargs):
        """Specifies additional QRDAMP modal analysis options.

        APDL Command: QRDOPT

        Parameters
        ----------
        reusekey
            Reuse key for method=QRDAMP specified in MODOPT command.

            ON
                Reuse the symmetric eigensolution from the previous
                load steps or from the previous solution.

            OFF
                Do not reuse (calculates symmetric eigensolution at
                current load step). This is the default.

        symmeth
            Mode-extraction method to be used for the symmetric eigenvalue
            problem.

            LANB
                Block Lanczos (default for shared-memory parallel processing).

            SUBSP
                Subspace algorithm (default for distributed-memory
                parallel processing).

        cmccoutkey
            Complex Modal Contribution Coefficients (CMCC) output
            key. See Calculate the Complex Mode Contribution
            Coefficients (CMCC) in the Structural Analysis Guide for
            details and usage.

            ON
                Output the CMCC to the text file Jobname.CMCC.

            OFF
                Do not output the CMCC. This is the default.

        Notes
        -----
        If the filename.modesym file exists in the working directory
        and ReuseKey = ON, filename.modesym will be reused. If
        filename.modesym does not exist in the working directory, the
        symmetric eigensolution will be calculated.

        When ReuseKey=ON, both the new modal analysis
        (filename.modesym usage) and the preceding modal analysis
        (filename.modesym generation) must be performed using the same
        product version number.

        The mode-extraction method changes depending on the type of
        parallelism involved. For performance reasons, the subspace
        method is used with distributed-memory parallel processing
        (Distributed ANSYS) runs, while the Block Lanczos method is
        used with shared-memory parallel processing runs.
        """
        return self.run(f"QRDOPT,{reusekey},,,{symmeth},{cmccoutkey}", **kwargs)

    def rigid(self, dof1="", dof2="", dof3="", dof4="", dof5="", dof6="", **kwargs):
        """Specifies known rigid body modes (if any) of the model.

        APDL Command: RIGID

        Parameters
        ----------
        dof1, dof2, dof3, . . . , dof6
            Up to six global Cartesian directions of the rigid modes.
            For a completely free 2-D model, use ALL or UX, UY, ROTZ.
            For a completely free 3-D model, use ALL or UX, UY, UZ,
            ROTX, ROTY, ROTZ.  For a constrained model, use UX, UY,
            UZ, ROTX, ROTY, or ROTZ, as appropriate, to specify each
            and every unconstrained direction which exists in the
            model (not specifying every direction may cause
            difficulties in extracting the modes).

        Notes
        -----
        Specifies known rigid body modes (if any) of the model.  This
        command applies only to a component mode synthesis (CMS)
        analysis (see the CMSOPT command).  Any rigid body modes
        specified must be permitted by the applied displacement
        constraints (i.e., do not specify a rigid body mode in a
        constrained direction).  Reissue the command to redefine the
        specification.  If used in SOLUTION, this command is valid
        only within the first load step.

        This command is also valid in PREP7.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RIGID,{dof1},{dof2},{dof3},{dof4},{dof5},{dof6}"
        return self.run(command, **kwargs)

    def subopt(self, option="", value1="", **kwargs):
        """Specifies Subspace (SUBSP) eigensolver options.

        APDL Command: SUBOPT

        Parameters
        ----------
        option
            One of the following options:

            STRMCK
                Controls whether a Sturm sequence check is performed.
                Valid input for Value1 when Option = STRMCK:

                OFF
                    Do not perform Sturm sequence check (default).

                ON
                    Perform Sturm sequence check.

            MEMORY
                Controls the memory allocation strategy for the Subspace eigensolver.
                Valid input for Value1 when Option = MEMORY:

                AUTO
                    Use the default memory allocation strategy (``default``).

                INCORE
                    Force the Subspace eigensolver to allocate in-core memory.

                MIX1
                    First level of mixed in-core / out-of-core configuration.

                MIX2
                    Second level of mixed in-core / out-of-core configuration.

                OUTOFCORE
                    Force the Subspace eigensolver to use scratch files.

            Value1
                Assigned value for the specified Option (as described above).

        Notes
        -----
        ``SUBOPT`` specifies options to be used with the Subspace eigensolver
        (``MODOPT,SUBSP``) during a modal analysis.
        """
        command = f"SUBOPT,{option},{value1}"
        return self.run(command, **kwargs)

    def timint(self, key="", lab="", **kwargs):
        """Turns on transient effects.

        APDL Command: TIMINT

        Parameters
        ----------
        key
            Transient effects key:

            OFF
                No transient effects (static or steady-state).

            ON
                Include transient (mass or inertia) effects.

        lab
            Degree of freedom label:

            ALL
                Apply this key to all appropriate labels (default).

            STRUC
                Apply this key to structural DOFs.

            THERM
                Apply this key to thermal DOFs.

            ELECT
                Apply this key to electric DOFs.

            MAG
                Apply this key to magnetic DOFs.

            FLUID
                Apply this key to fluid DOFs.

            DIFFU
                Apply this key to concentration of DOFs.

        Notes
        -----
        Indicates whether this load step in a full transient analysis should
        use time integration, that is, whether it includes transient effects
        (e.g. structural inertia, thermal capacitance) or whether it is a
        static (steady-state) load step for the indicated DOFs.  Transient
        initial conditions are introduced at the load step having Key = ON.
        Initial conditions are then determined from the previous two substeps.
        Zero initial velocity and acceleration are assumed if no previous
        substeps exist.  See the Structural Analysis Guide, the Thermal
        Analysis Guide, and the Low-Frequency Electromagnetic Analysis Guide
        for details.

        This command is also valid in PREP7.
        """
        command = f"TIMINT,{key},{lab}"
        return self.run(command, **kwargs)

    def tintp(
        self,
        gamma="",
        alpha="",
        delta="",
        theta="",
        oslm="",
        tol="",
        avsmooth="",
        alphaf="",
        alpham="",
        **kwargs,
    ):
        """Defines transient integration parameters.

        APDL Command: TINTP

        Parameters
        ----------
        gamma
            Amplitude decay factor for 2nd order transient integration, e.g.,
            structural dynamics (used only if ALPHA, DELTA, ALPHAF, and ALPHAM
            are blank).  Defaults to 0.005.

        alpha
            2nd order transient integration parameter (used only if GAMMA is
            blank).  Defaults to 0.2525.

        delta
            2nd order transient integration parameter (used only if GAMMA is
            blank).  Defaults to 0.5050.

        theta
            1st order transient (e.g., thermal transient) integration
            parameter.  Defaults to 1.0.

        oslm
            Specifies the oscillation limit criterion for automatic time
            stepping of 1st order transients (e.g., thermal transients).
            Defaults to 0.5 with a tolerance of TOL.

        tol
            Tolerance applied to OSLM.  Defaults to 0.0.

        avsmooth
            Smoothing flag option:

            0
                Include smoothing of the velocity (1st order system)
                or the acceleration (2nd order system) (default).

            1
                Do not include smoothing.

        alphaf
            Interpolation factor in HHT algorithm for force and damping terms
            (used only if GAMMA is blank). Defaults to 0.005.

        alpham
            Interpolation factor in HHT algorithm for inertial term (used only
            if GAMMA is blank). Defaults to 0.0.

        Notes
        -----
        Used to define the transient integration parameters.  For more
        information on transient integration parameters, refer to the
        Mechanical APDL Theory Reference.

        For structural transient analyses, you may choose between the
        Newmark and HHT time integration methods (see the TRNOPT
        command). In this case, if GAMMA is input and the integration
        parameters ALPHA, DELTA, ALPHAF, and ALPHAM are left blank,
        the program will calculate the integration
        parameters. Alternatively, you can input these integration
        parameters directly on this command. However, for the
        unconditional stability and second order accuracy of the time
        integration, these parameters should satisfy a specific
        relationship, as described in Description of Structural and
        Other Second Order Systems of the Mechanical APDL Theory
        Reference.

        In a transient piezoelectric analysis, required input for this
        command is ALPHA = 0.25, DELTA = 0.5, and THETA = 0.5.  For a
        coupled electromagnetic-circuit transient analysis, use THETA
        = 1.0, the default value, to specify the backward Euler
        method.

        This command is also valid in PREP7.
        """
        command = f"TINTP,{gamma},{alpha},{delta},{theta},{oslm},{tol},,,{avsmooth},{alphaf},{alpham}"
        return self.run(command, **kwargs)

    def trnopt(
        self,
        method="",
        maxmode="",
        initialacc="",
        minmode="",
        mcfwrite="",
        tintopt="",
        vaout="",
        dmpsfreq="",
        engcalc="",
        mckey="",
        **kwargs,
    ):
        """Specifies transient analysis options.

        APDL Command: TRNOPT

        Parameters
        ----------
        method
            Solution method for the transient analysis:

            FULL
                Full method (default).

            MSUP
                Mode-superposition method.

            VT
                Variational Technology method.  (Removed by V18.2)

        maxmode
            Largest mode number to be used to calculate the response
            (for Method = MSUP).  Defaults to the highest mode
            calculated in the preceding modal analysis.

        initialacc
            Key to activate calculation of initial acceleration:

            (blank)
                Initial accelerations are not calculated (default).

            INIL
                Calculate initial acceleration for a full transient analysis
                using the lumped mass matrix.

        minmode
            Smallest mode number to be used (for Method = MSUP).
            Defaults to 1.

        mcfwrite
            Modal coordinates output key to the .mcf file (valid only
            for the mode-superposition method):

            NO
                No output of modal coordinates (default).

            YES
                Output modal coordinates to the text file Jobname.MCF.

        tintopt
            Time integration method for the transient analysis:

            NMK or 0
                Newmark algorithm (default).

            HHT or 1
                HHT algorithm (valid only for the full transient method).

        vaout
            Velocities and accelerations output key (valid only for
            mode- superposition transient analysis):

            NO
                No output of velocities and accelerations (default).

            YES
                Write velocities and accelerations on the reduced displacement
                file Jobname.RDSP.

        dmpsfreq
            Average excitation frequency (Hz) for the calculation of
            equivalent viscous damping from structural damping input
            (DMPSTR and MP,DMPS). See Damping for more
            details. Defaults to zero. If an excitation frequency is
            not specified, structural damping is ignored. If tabular
            excitation frequency data is provided in a full transient
            analysis (DMPSFreqTab on DMPSTR), it supersedes this
            value.

        engcalc
            Additional element energies calculation key:

            NO
                Do not calculate additional element energies
                (default).

            YES
                Calculate damping energy and work done by external
                loads.

        mckey
            Modal coordinates output key to the .rdsp file (valid only
            for the mode-superposition method):

            AUTO
                Writing depends on the modal analysis settings of
                the MXPAND command (default).

            YES
                Always write the modal coordinates to the file
                Jobname.rdsp. A subsequent expansion pass (EXPASS) is not
                supported.

        Notes
        -----
        Specifies transient analysis (ANTYPE,TRANS) options. If used
        in SOLUTION, this command is valid only within the first load
        step. Use the TINTP command to set transient integration
        parameters.

        To include residual vectors in your mode-superposition
        transient analysis (Method = MSUP), specify RESVEC,ON.

        Method = MSUP is not available for ocean loading.

        By default in a mode-superposition transient analysis,
        reaction force and other force output contains only static
        contributions. If you want to postprocess the velocities,
        accelerations, and derived results (Lab = TOTAL, DAMP, or
        INERT on the FORCE command), set VAout = YES to activate
        velocity and acceleration output.

        The calculation of additional energies (EngCalc = YES) is
        valid only for the full solution method (Method = FULL). The
        Jobname.ESAV file is always saved in this case. The numerical
        integration for damping energy and work are consistent only if
        solution data are written to the database for every substep
        (OUTRES,ALL,ALL, OUTRES,ESOL,ALL, or OUTRES,VENG, ALL). For
        more information, see Damping Energy and Work Done by External
        Loads in the Mechanical APDL Theory Reference.

        This command is also valid in PREP7.
        """
        command = f"TRNOPT,{method},{maxmode},{initialacc},{minmode},{mcfwrite},{tintopt},{vaout},{dmpsfreq},{engcalc},{mckey}"
        return self.run(command, **kwargs)
        # TRNOPT,Method,MAXMODE,InitialAcc,MINMODE,MCFwrite,TINTOPT,VAout, DMPSFreq, EngCalc, MCkey



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Contains the Node implementation of the MapdlDb class.

This allows access to the Nodes in the MAPDL DB from Python.

"""
import weakref

from ansys.api.mapdl.v0 import ansys_kernel_pb2 as anskernel
from ansys.api.mapdl.v0 import mapdl_db_pb2
import numpy as np
from numpy.lib import recfunctions

from ansys.mapdl.core.errors import MapdlRuntimeError

from ..common_grpc import DEFAULT_CHUNKSIZE
from .database import DBDef, MapdlDb, check_mapdl_db_is_alive


class DbNodes:
    """
    Abstract mapdl db nodes class.

    Examples
    --------
    Create a nodes instance.

    >>> from ansys.mapdl.core import launch_mapdl
    >>> mapdl = launch_mapdl()
    >>> # create nodes...
    >>> nodes = mapdl.db.nodes
    >>> print(nodes)
    MAPDL Database Nodes
        Number of nodes:          270641
        Number of selected nodes: 270641
        Maximum node number:      270641

    >>> mapdl.nsel("NONE")
    >>> print(nodes)
    MAPDL Database Nodes
        Number of nodes:          270641
        Number of selected nodes: 0
        Maximum node number:      270641

    Return the selection status and the coordinates of node 22.

    >>> nodes = mapdl.db.nodes
    >>> sel, coord = nodes.coord(22)
    >>> coord
    (1.0, 0.5, 0.0, 0.0, 0.0, 0.0)

    Return all the node indices, coordinates, and angles of all the nodes.

    >>> nodes = mapdl.db.nodes
    >>> ind, coords, angles = nodes.all_asarray()
    >>> ind
    array([     1,      2,      3, ..., 270639, 270640, 270641], dtype=int32)

    >>> coords
    array([[0.    , 1.    , 0.    ],
           [0.    , 0.    , 0.    ],
           [0.    , 0.9875, 0.    ],
           ...,
           [0.9875, 0.975 , 0.925 ],
           [0.9875, 0.975 , 0.95  ],
           [0.9875, 0.975 , 0.975 ]])

    >>> angles
    array([[0., 0., 0.],
           [0., 0., 0.],
           [0., 0., 0.],
           ...,
           [0., 0., 0.],
           [0., 0., 0.],
           [0., 0., 0.]])

    """

    def __init__(self, db):
        """Initialize this class."""
        if not isinstance(db, MapdlDb):  # pragma: no cover
            raise TypeError("``db`` must be a MapdlDb instance")
        self._db_weakref = weakref.ref(db)
        self._itnod = -1

    def __str__(self):
        """Return the string representation of this class."""
        lines = ["MAPDL Database Nodes"]
        lines.append(f"    Number of nodes:          {self.num()}")
        lines.append(f"    Number of selected nodes: {self.num(selected=True)}")
        lines.append(f"    Maximum node number:      {self.max_num}")
        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def _db(self):
        """Return the weakly referenced instance of db."""
        return self._db_weakref()

    @check_mapdl_db_is_alive
    def first(self, inod=0):
        """
        Return the number of the first node.

        This starts at ``inod``, defaults to the first node in the model.

        Parameters
        ----------
        inod : int, optional
            The first node number to consider as the "first node".

        Returns
        -------
        int
            The first node number within either selected or all nodes.

        Examples
        --------
        Return the first selected node.

        >>> nodes.first()
        1

        Return the first node after node 10.

        >>> nodes.first(inod=10)
        11

        """
        self._itnod = inod
        return self.next()

    @check_mapdl_db_is_alive
    def next(self):
        """
        Return the number of the next selected node.

        You must first call :func:`DbNodes.first`.

        Returns
        -------
        int
            The next selected node number. Returns 0 if there are no more nodes.

        Examples
        --------
        Call :func:`DbNodes.first` first.

        >>> nodes.first()
        1

        Then get the next node.

        >>> nodes.next()
        2

        """
        if self._itnod == -1:
            raise MapdlRuntimeError(
                "You first have to call the `DbNodes.first` method."
            )

        request = mapdl_db_pb2.NodRequest(next=self._itnod)
        result = self._db._stub.NodNext(request)
        self._itnod = result.inum
        return self._itnod

    # def next_defined(self):
    #     """get the number of the next defined node
    #     You first have to call first)_

    #     Returns
    #     -------
    #     next_defined : int
    #     The next defined node number
    #     = 0 - no more nodes
    #     """

    #     if self._itnod == -1:
    #         raise TypeError(
    #             "``db.next_node`` you first have to call first_node function"
    #         )
    #     breakpoint()
    #     request = mapdl_db_pb2.NodRequest(next=self._itnod)
    #     result = self._db._stub.NodNextDefined(request)
    #     self._itnod = result.inum
    #     return self._itnod

    @check_mapdl_db_is_alive
    def info(self, inod, ikey):
        """
        Return information about a node.

        Parameters
        ----------
        inod : int
            Node number. Should be 0 for key=11 for the following:

            * ``DB_NUMDEFINED``
            * ``DB_NUMSELECTED``
            * ``DB_MAXDEFINED``
            * ``DB_MAXRECLENG``

        ikey : int
            Key of the information needed about the node. One of the following:

            * DB_SELECTED : return select status

                * 0 - node is undefined.
                * -1 - node is unselected.
                *  1 - node is selected.

            * DB_NUMDEFINED - return number of defined nodes
            * DB_NUMSELECTED - return number of selected nodes
            * DB_MAXDEFINED - return highest node number defined
            * DB_MAXRECLENG - return maximum record length (dp words)
            * 2 - length (dp words)
            * 3 -
            * 4 - pointer to first data word
            * 11 - return void percent (integer)
            * 17 - pointer to start of index
            * 117 - return the maximum number of DP contact data stored for any node
            * -1 -
            * -2 - superelement flag
            * -3 - master dof bit pattern
            * -4 - active dof bit pattern
            * -5 - solid model attachment
            * -6 - pack nodal line parametric value
            * -7 - constraint bit pattern
            * -8 - force bit pattern
            * -9 - body force bit pattern
            * -10 - internal node flag
            * -11 - orientation node flag =1 is =0 isnot
            * -11 - contact node flag <0
            * -12 - constraint bit pattern (for DSYM)
            * -13 - if dof constraint written to file. (for LSDYNA only)
            * -14 - nodal coordinate system number (set by NROTATE)
            * -101 - pointer to node data record
            * -102 - pointer to angle record
            * -103 -
            * -104 - pointer to attached couplings
            * -105 - pointer to attacted constraint equations
            * -106 - pointer to nodal stresses
            * -107 - pointer to specified disp'S
            * -108 - pointer to specified forces
            * -109 - pointer to x/y/z record
            * -110 -
            * -111 -
            * -112 - pointer to nodal temperatures
            * -113 - pointer to nodal heat generations
            * -114 -
            * -115 - pointer to calculated displacements
            * -116 -

        Returns
        -------
        int
            Information from the query based on ``ikey``.

        Examples
        --------
        Query if a node is selected.

        >>> from ansys.mapdl.core.database import DBDef
        >>> nodes = mapdl.db.nodes
        >>> nodes.info(1, DBDef.DB_SELECTED)
        1

        """
        if isinstance(ikey, DBDef):
            ikey = ikey.value
        request = mapdl_db_pb2.NodInqrRequest(node=inod, key=ikey)
        result = self._db._stub.NodInqr(request)
        return result.ret

    @check_mapdl_db_is_alive
    def num(self, selected=False) -> int:
        """
        Return the number of nodes, either selected or all.

        Parameters
        ----------
        selected : bool, optional
            Return either the number of selected nodes of the total number of
            nodes.

        Returns
        -------
        int
            Number of nodes.

        Examples
        --------
        Get the number of selected nodes.

        >>> nodes = mapdl.db.nodes
        >>> nodes.num(selected=True)
        425

        """
        if selected:
            return self.info(0, DBDef.DB_NUMSELECTED.value)
        return self.info(0, DBDef.DB_NUMDEFINED.value)

    @property
    @check_mapdl_db_is_alive
    def max_num(self) -> int:
        """
        Return the maximum node number.

        Examples
        --------
        Return the maximum node number.

        >>> nodes = mapdl.db.nodes
        >>> nodes.max_num
        425

        """
        return self.info(0, DBDef.DB_MAXDEFINED.value)

    @check_mapdl_db_is_alive
    def coord(self, inod):
        """
        Return the location of a node.

        Parameters
        ----------
        inod : int
            Node number.

        Returns
        -------
        int
            Selection status:

            * 0 - node is selected
            * 1 - node is not defined
            * -1 - node is unselected

        tuple
            Coordinates (first 3 values) and rotation angles (last 3 values).

        Examples
        --------
        Return the selection status and the coordinates of node 22.

        >>> nodes = mapdl.db.nodes
        >>> sel, coord = nodes.coord(22)
        >>> coord
        (1.0, 0.5, 0.0, 0.0, 0.0, 0.0)

        """
        request = mapdl_db_pb2.getnodRequest(node=inod)
        node = self._db._stub.getNod(request)
        return node.kerr, tuple(node.v)

    @check_mapdl_db_is_alive
    def all_asarray(self):
        """
        Return all node indices, coordinates, and angles as arrays.

        .. note::
           This only returns data of the selected nodes.

        Returns
        -------
        np.ndarray
            Numpy.int32 array of node numbers with shape ``(n_node,)``.
        np.ndarray
            Numpy.double array of node coordinates with shape ``(n_node, 3)``.
        np.ndarray
            Numpy.double array of node angles with shape ``(n_node, 3)``.

        Examples
        --------
        Return all the node indices, coordinates, and angles of all the nodes.

        >>> nodes = mapdl.db.nodes
        >>> ind, coords, angles = nodes.all_asarray()
        >>> ind
        array([     1,      2,      3, ..., 270639, 270640, 270641], dtype=int32)

        >>> coords
        array([[0.    , 1.    , 0.    ],
               [0.    , 0.    , 0.    ],
               [0.    , 0.9875, 0.    ],
               ...,
               [0.9875, 0.975 , 0.925 ],
               [0.9875, 0.975 , 0.95  ],
               [0.9875, 0.975 , 0.975 ]])

        >>> angles
        array([[0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.],
               ...,
               [0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.]])

        """
        chunk_size = DEFAULT_CHUNKSIZE
        metadata = [("chunk_size", str(chunk_size))]
        request = anskernel.EmptyRequest()
        chunks = self._db._stub.getAllNodC(request, metadata=metadata)

        n_nodes = int(chunks.initial_metadata()[0][1])

        ind = np.empty(n_nodes, np.int32)
        coord = np.empty((n_nodes, 3), np.double)
        angle = np.empty((n_nodes, 3), np.double)

        # each "chunk" from MAPDL is organized as:
        # n_nodes, node, node ... last_node
        #
        # where each node is
        # INT32, DOUBLE, DOUBLE, DOUBLE, DOUBLE, DOUBLE, DOUBLE

        # first, just interpret as int32 and extract node numbering
        # ignore the first int as this is n_nodes
        # data = parse_chunks(chunks, np.int32)
        c = 0
        for chunk in chunks:
            parsed_chunk = np.frombuffer(chunk.payload, np.int32)
            n_node = parsed_chunk[0]
            data = parsed_chunk[1:].view(
                [
                    ("ind", np.int32),
                    ("x", np.double),
                    ("y", np.double),
                    ("z", np.double),
                    ("xang", np.double),
                    ("yang", np.double),
                    ("zang", np.double),
                ]
            )

            ind_chunk = recfunctions.structured_to_unstructured(data[["ind"]]).ravel()

            coord_chunk = recfunctions.structured_to_unstructured(
                data[["x", "y", "z"]]
            ).reshape((-1, 3))

            angle_chunk = recfunctions.structured_to_unstructured(
                data[["xang", "yang", "zang"]]
            ).reshape((-1, 3))

            ind[c : c + n_node] = ind_chunk
            coord[c : c + n_node] = coord_chunk
            angle[c : c + n_node] = angle_chunk
            c += n_node

        return ind, coord, angle

    @check_mapdl_db_is_alive
    def push(self, inod, x, y, z, xang=None, yang=None, zang=None):
        """
        Push a single node into the DB.

        Parameters
        ----------
        inod : int
            Node number.
        x : float
            X coordinate.
        y : float
            X coordinate.
        z : float
            Z coordinate.
        xang : float, optional
            X angle.
        yang : float, optional
            X angle.
        zang : float, optional
            Z angle.

        Examples
        --------
        Update node 1 to have coordinates ``(10.0, 20.0, 30.0)``.

        >>> nodes = mapdl.db.nodes
        >>> nodes.push(1, 10, 20, 30)
        >>> sel, coord = nodes.coord(1)
        >>> coord
        (10.0, 20.0, 30.0, 0.0, 0.0, 0.0)

        """
        request = mapdl_db_pb2.putnodRequest()
        request.node = inod
        request.vctn.extend([x, y, z])
        if xang:
            request.vctn.extend([xang])
        if yang:
            if xang is None:
                raise ValueError("X angle must be input as well when inputing Y angle")
            request.vctn.extend([yang])
        if zang:
            if xang is None or yang is None:
                raise ValueError(
                    "X and Y angles must be input as well when inputing Z angle"
                )
            request.vctn.extend([zang])

        self._db._stub.putNod(request)

    ###############################################################################
    # unimplemented

    # def _select(self, inod, sel=1):
    #     """Select, unselect, or delete a node in the database.

    #     Parameters
    #     ----------
    #     inod : int
    #         Node number.
    #     sel : int
    #         Action to take. One of the following:

    #         * ``0`` - Delete the node.
    #         * ``1`` - Select a node.
    #         * ``-1`` - Unselect a node.
    #         * ``2`` - Switch the exist select status.

    #     """
    #     request = mapdl_db_pb2.NodSelRequest(inum=inod, ksel=sel)
    #     self._db._stub.NodSel(request)

    # def delete(self, inod):
    #     """Delete a node from the database.

    #     Parameters
    #     ----------
    #     inod : int
    #         Node number.

    #     Examples
    #     --------
    #     Push and then delete a node.

    #     >>> nodes = mapdl.db.nodes
    #     >>> nodes.push(1, 10, 20, 30)
    #     >>> sel, coord = nodes.coord(1)
    #     >>> nodes.delete(1)

    #     """
    #     return self._select(inod, 0)

    # def select(self, inod):
    #     """select a node in the database.

    #     Parameters
    #     ----------
    #     inod : int
    #         Node number.

    #     Examples
    #     --------

    #     """
    #     self._select(inod, 1)

    # def unselect(self, inod):
    #     """Unselect a node in the database.

    #     Parameters
    #     ----------
    #     inod : int
    #         Node number.

    #     Examples
    #     --------

    #     """
    #     self._select(inod, -1)

    # def invselect(self, inod):
    #     """Inverse the select status of a node in the database.

    #     Parameters
    #     ----------
    #     inod : int
    #         Node number.

    #     Examples
    #     --------

    #     """
    #     self._select(inod, 2)

    ###############################################################################
